// Copyright (C) 2012 David Sugar, Tycho Softworks.
//
// This file is part of  cape-devtools.
//
// Cape-devtools is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Cape-devtools is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Cape-devtools.    If not, see <http://www.gnu.org/licenses/>.

#include <cape-devtools-config.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/wait.h>
#include <sys/utsname.h>
#include <sys/fcntl.h>
#include <unistd.h>
#include <pwd.h>
#include <ucommon/ucommon.h>

using namespace UCOMMON_NAMESPACE;

class __LOCAL plock
{
public:
    char *path;
    pid_t pid;

    inline operator bool()
        {return path != NULL;};

    inline bool operator!()
        {return path == NULL;};

    plock(char *lockfile, timeout_t timer);
    ~plock();

    void release(void);
};

static shell::flagopt helpflag('h',"--help",    _TEXT("display this list"));
static shell::flagopt althelp('?', NULL, NULL);
static shell::stringopt distid('d', "--distribution", _TEXT("alternate distribution name to use"), "id", NULL);
static shell::flagopt ignore('i', "--ignore-index", _TEXT("dont create index"));
static shell::flagopt preserve('k', "--keep", _TEXT("keep prior releases"));
static shell::flagopt move('m', "--move", _TEXT("remove source after copy"));
static shell::stringopt profile('p', "--profile", _TEXT("project profile"), "config", "default");
static shell::flagopt sign('s', "--sign", _TEXT("sign index"));
static shell::flagopt verbose('v', "--verbose", _TEXT("verbose"));
static bool debianos = false, redhatos = false;
static const char *signid = NULL, *origin = "localhost", *label = "", *distribution = NULL;
static const char *email = NULL, *realname = NULL;
static keyfile pkgbuildrc;

static void copyfile(String& from, String& to)
{
    char buffer[4096];

    fd_t source, target;
    fsys::fileinfo_t finfo, tinfo;
    String tmp = str(to) + ".tmp.";
    if(fsys::info(*from, &finfo))
        return;

    // if file unchanged, no need to copy
    if(!fsys::info(*to, &tinfo)) {
        if(finfo.st_mtime < tinfo.st_mtime && finfo.st_size == tinfo.st_size)
            return;
    }

    source = open(*from, O_RDONLY);
    if(source < 0)
        return;

    remove(*tmp);

    // if another session already doing it, ignore...
    target = open(*tmp, O_CREAT | O_EXCL | O_WRONLY, finfo.st_mode & 0777);
    if(target < 0) {
        close(source);
        return;
    }

    const char *base = strrchr(*to, '/');
    if(!base)
        base = *to;
    else
        ++base;
    if(is(verbose))
        printf("copying %s\n", base);

    for(;;) {
        ssize_t len = read(source, buffer, sizeof(buffer));
        if(len < 0) {
            close(source);
            close(target);
            remove(*tmp);
            fprintf(stderr, "*** cape-archive: %s: %s\n",
                *from, _TEXT("unable to copy"));
        }
        if(len > 0)
            if(write(target, buffer, len) < len)
                shell::errexit(1, "*** cape-archive: %s",
                    _TEXT("cannot copy"));
        if((size_t)len < sizeof(buffer))
            break;
    }
    close(source);
    close(target);
    rename(*tmp, *to);
}

static void erasefiles(const char *path, const char *prefix, const char *suffix)
{
    // option to preserve old releases...
    if(is(preserve))
        return;

    char buffer[65];
    dir_t dir(path);
    size_t plen = strlen(prefix), slen = strlen(suffix), len;
    while(is(dir) && dir.read(buffer, sizeof(buffer)) > 0) {
        len = strlen(buffer);
        if(len <= slen)
            continue;
        if(!eq(prefix, buffer, plen))
            continue;
        if(!eq(suffix, buffer + len - slen))
            continue;
        if(is(verbose))
            printf("removing %s\n", buffer);
        fsys::erase(str(path) + "/" + (const char *)buffer);
    }
}

static int bz2(const char *source, const char *target)
{
    char *argv[] = {(char *)"bzip2", NULL};

    pid_t pid = fork();
    if(pid)
        return shell::wait(pid);
    else {
        fd_t fd = open(target, O_TRUNC | O_CREAT | O_WRONLY, 0755);
        dup2(fd, 1);
        close(fd);
        fd = open(source, O_RDONLY);
        dup2(fd, 0);
        execvp(*argv, argv);
        exit(-1);
    }
}

static int gzip(const char *source, const char *target)
{
    char *argv[] = {(char *)"gzip", NULL};

    pid_t pid = fork();
    if(pid)
        return shell::wait(pid);
    else {
        fd_t fd = open(target, O_TRUNC | O_CREAT | O_WRONLY, 0755);
        dup2(fd, 1);
        close(fd);
        fd = open(source, O_RDONLY);
        dup2(fd, 0);
        execvp(*argv, argv);
        exit(-1);
    }
}

static int dsc_index(const char *path, const char *subdir)
{
    if(chdir(path))
        shell::errexit(10, "*** cape-archive: %s: %s",
            path, _TEXT("invalid directory"));
    unsigned status = 0;
    pid_t pid;
    String output = str(subdir) + "/Sources";
    String gzoutput = str(subdir) + "/Sources.gz";
    String bz2output = str(subdir) + "/Sources.bz2";

    char *scanv[] = {(char *)"dpkg-scansources", (char *)subdir, NULL};

    pid = fork();
    if(pid)
        status = shell::wait(pid);
    else {
        fd_t fd = open(*output, O_TRUNC | O_CREAT | O_WRONLY, 0755);
        dup2(fd, 1);
        close(fd);
        char **argv = scanv;
        execvp(*argv, argv);
        exit(-1);
    }
    if(status)
        return status;

    status = gzip(*output, *gzoutput);
    if(status)
        return status;

    return bz2(*output, *bz2output);
}

static int deb_index(const char *path, const char *subdir)
{
    if(chdir(path))
        shell::errexit(10, "*** cape-archive: %s: %s",
            path, _TEXT("invalid directory"));
    unsigned status = 0;
    pid_t pid;
    String output = str(subdir) + "/Packages";
    String gzoutput = str(subdir) + "/Packages.gz";
    String bz2output = str(subdir) + "/Packages.bz2";

    char *scanv[] = {(char *)"dpkg-scanpackages", (char *)"-m", (char *)subdir, NULL};

    pid = fork();
    if(pid)
        status = shell::wait(pid);
    else {
        fd_t fd = open(*output, O_TRUNC | O_CREAT | O_WRONLY, 0755);
        dup2(fd, 1);
        close(fd);
        char **argv = scanv;
        execvp(*argv, argv);
        exit(-1);
    }
    if(status)
        return status;

    status = gzip(*output, *gzoutput);
    if(status)
        return status;

    return bz2(*output, *bz2output);
}

static int deb_release(const char *source, const char *subdir)
{
    String release = str(subdir) + "/Release";
    FILE *fp = fopen(*release, "w"), *md;
    static const char *files[] = {"Packages", "Packages.gz", "Packages.bz2", "Sources", "Sources.gz", "Sources.bz2", NULL};
    unsigned index;
    const char *cp;
    char *ep;
    fsys::fileinfo_t ino;
    char cmd[128];
    String mdfile;
    bool md5header = false, sha1header = false, sha256header = false;

    const char *distname = subdir;
    if(distribution)
        distname = distribution;

    fprintf(fp, "Archive: %s\n", distname);
    fprintf(fp, "Codename: %s\n", distname);
    fprintf(fp, "Origin: %s\n", origin);
    fprintf(fp, "Label: %s\n", label);
    fprintf(fp, "Architecture: all\n");

    index = 0;
    while(NULL != (cp = files[index++])) {
        mdfile = str(subdir) + "/" + cp;
        if(fsys::info(*mdfile, &ino))
            continue;

        snprintf(cmd, sizeof(cmd), "md5sum %s", *mdfile);
        md = popen(cmd, "r");
        if(!md)
            continue;
        if(!fgets(cmd, sizeof(cmd), md)) {
            pclose(md);
            continue;
        }

        if(!md5header) {
            fprintf(fp, "MD5Sum:\n");
            md5header = true;
        }

        ep = strchr(cmd, ' ');
        if(ep)
            *ep = 0;
        else
            cmd[0] = 0;
        fprintf(fp, " %s %lu %s\n", cmd, ino.st_size, cp);
        pclose(md);
    }

    index = 0;
    while(NULL != (cp = files[index++])) {
        mdfile = str(subdir) + "/" + cp;
        if(fsys::info(*mdfile, &ino))
            continue;

        snprintf(cmd, sizeof(cmd), "sha1sum %s", *mdfile);
        md = popen(cmd, "r");
        if(!md)
            continue;
        if(!fgets(cmd, sizeof(cmd), md)) {
            pclose(md);
            continue;
        }

        if(!sha1header) {
            fprintf(fp, "SHA1:\n");
            sha1header = true;
        }

        ep = strchr(cmd, ' ');
        if(ep)
            *ep = 0;
        else
            cmd[0] = 0;
        fprintf(fp, " %s %lu %s\n", cmd, ino.st_size, cp);
        pclose(md);
    }

    index = 0;
    while(NULL != (cp = files[index++])) {
        mdfile = str(subdir) + "/" + cp;
        if(fsys::info(*mdfile, &ino))
            continue;

        snprintf(cmd, sizeof(cmd), "sha256sum %s", *mdfile);
        md = popen(cmd, "r");
        if(!md)
            continue;
        if(!fgets(cmd, sizeof(cmd), md)) {
            pclose(md);
            continue;
        }

        if(!sha256header) {
            fprintf(fp, "SHA256:\n");
            sha256header = true;
        }

        ep = strchr(cmd, ' ');
        if(ep)
            *ep = 0;
        else
            cmd[0] = 0;
        fprintf(fp, " %s %lu %s\n", cmd, ino.st_size, cp);
        pclose(md);
    }
    fclose(fp);

    const char *out = strdup(str(release) + ".gpg");
    fsys::erase(out);

    char *argv[16];
    index = 0;
    argv[index++] = (char *)"gpg";
    if(signid) {
        argv[index++] = (char *)"-u";
        argv[index++] = (char *)signid;
    }
    argv[index++] = (char *)"-ba";
    argv[index++] = (char *)"-o";
    argv[index++] = (char *)out;
    argv[index++] = (char *)"--sign";
    argv[index++] = (char *)*release;
    argv[index++] = NULL;

    return fsys::exec(*argv, argv);
}

static int rpm_index(const char *path)
{
    char *cv[] = {(char *)"createrepo", (char *)"--update", (char *)path, NULL};

    char **argv = cv;
    return fsys::exec(*argv, argv);
}

static int rpm_release(const char *prefix)
{
    String path = str(prefix) + "/repodata/repomd.xml";
    char *argv[16];
    unsigned index = 0;

    argv[index++] = (char *)"gpg";
    if(signid) {
        argv[index++] = (char *)"-u";
        argv[index++] = (char *)signid;
    }
    argv[index++] = (char *)"--detach-sign";
    argv[index++] = (char *)"--armor";
    argv[index++] = (char *)*path;
    argv[index++] = NULL;

    return fsys::exec(*argv, argv);
}

static void deb_update(const char *dir, const char *base, const char *archive)
{
    String fname;
    String from = str(dir) + "/" + base;
    String to = str(archive) + "/" + base;
    file_t changes(*from, "r");
    bool files = false;
    char *fn, *fe, *fext;
    char *version = NULL, *revision = NULL;
    char buffer[256];
    const char *ext = strrchr(base, '.');
    char *ep = (char *)strchr(base, '_');
    char *lp = (char *)strrchr(base, '_');
    ep = strchr((char *)base, '_');
    if(ep)
        *ep = 0;

    while(changes.getline(buffer, sizeof(buffer))) {
        if(eq(buffer, "Version:", 8)) {
            version = buffer + 8;
            while(isspace(*version))
                ++version;
            version = strdup(version);
            revision = strchr(version, '-');
            if(revision)
                *(revision++) = 0;
            continue;
        }
        if(eq(buffer, "Files:", 6)) {
            files = true;
            continue;
        }
        if(buffer[0] != ' ' || !files) {
            files = false;
            continue;
        }
        if(version == NULL)
            continue;

        fn = strchr(buffer + 2, ' ');
        while(NULL != (fe = strchr(++fn, ' '))) {
            fn = fe;
        }
        fname = str(fn);
        fe = strchr(fn, '_');
        if(!fe)
            continue;
        *(fe++) = 0;
        fext = strstr(fe, ".orig.");
        if(!fext)
            fext = strchr(fe, '_');
        if(!fext && revision)
            fext = fe + strlen(version) + strlen(revision) + 1;
        else if(!fext)
            fext = fe + strlen(version);

        erasefiles(archive, str(fn) + "_", fext);
        String ffrom = str(dir) + "/" + fname;
        String fto = str(archive) + "/" + fname;
        copyfile(ffrom, fto);
        if(is(move))
            fsys::erase(*ffrom);
    }

    if(ep != lp)
        ext = lp;
    erasefiles(archive, str(base) + "_", ext);
    copyfile(from, to);
    if(is(move))
        fsys::erase(*from);
}

static void rpm_update(const char *dir, const char *base, const char *archive)
{
    String from = str(dir) + "/" + base;
    String to = str(archive) + "/" + base;
    char *ep = (char *)base;
    char *dash = NULL;

    while(*ep) {
        if(*ep == '-') {
            dash = ep++;
            continue;
        }
        if(dash && *ep >= '0' && *ep <= '9') {
            *dash = 0;
            break;
        }
        ++ep;
    }
    ep = strrchr(ep, '.');
    while(*(--ep) != '.')
        continue;

    erasefiles(archive, str(base) + "-", ep);
    copyfile(from, to);
}

static void tar_update(const char *dir, const char *base, const char *version, const char *ext, const char *archive)
{
    char *argv[16];
    unsigned index = 0;

    String from = str(dir) + "/" + base + "-" + version + "." + ext;
    String to = str(archive) + "/" + base + "-" + version + "." + ext;

    erasefiles(archive, str(base) + "-", ext);
    erasefiles(archive, str(base) + "-", str(ext) + ".sig");
    erasefiles(archive, str(base) + "-", str(ext) + ".directive");
    copyfile(from, to);

    if(is(sign)) {
        index = 0;
        argv[index++] = (char *)"gpg";
        argv[index++] = (char *)"-b";
        argv[index++] = (char *)"-o";
        argv[index++] = strdup(str(to) + ".sig");

        if(signid) {
            argv[index++] = (char *)"-u";
            argv[index++] = (char *)signid;
        }
        argv[index++] = strdup(*to);
        argv[index++] = 0;
        if(fsys::exec(*argv, argv))
            shell::errexit(3, "*** cape-archive: %s: %s",
                *to, _TEXT("unable to sign"));
    }
}

PROGRAM_MAIN(argc, argv)
{
    char hostname[256];
    char filename[65];
    char prefix[256];
    char *ext;
    const char *archive;

    if(fsys::is_file("/etc/debian_version"))
        debianos = true;

    if(fsys::is_dir("/etc/yum"))
        redhatos = true;

    shell::bind("cape-archive");
    shell args(argc, argv);
    argv = args.argv();
    argc = args.argc();

    struct passwd *pwd = getpwuid(getuid());

    pkgbuildrc.load(shell::path(shell::USER_HOME) + "/.pkgbuildrc");
    keydata *archives = pkgbuildrc.get(*profile);
    gethostname(hostname, sizeof(hostname));

    if(pwd) {
        email = strdup(str(pwd->pw_name) + "@" + (const char *)hostname);
        realname = strdup(pwd->pw_gecos);
        char *ep = strchr((char *)realname, ',');
        if(ep)
            *ep = 0;
    }
    else {
        email = strdup(str(getlogin()) + "@" + (const char *)hostname);
        realname = strdup(getlogin());
    }

    if(archives) {
        label = archives->get("label");
        origin = archives->get("origin");
        distribution = archives->get("distribution");
        const char *em = archives->get("email");
        signid = archives->get("key");
        if(!signid)
            signid = em;
        if(em)
            email = em;
    }

    if(is(distid))
        distribution = *distid;

    setenv("DEBEMAIL", email, 1);
    setenv("DEBFULLNAME", realname, 1);

    if(is(helpflag) || is(althelp) || argc < 1) {
        printf("%s: cape-archive %s\n", _TEXT("Usage"),
            _TEXT("[options] [pkgs] target-archive-dir"));
        printf("%s\n\n", _TEXT("Archive packages and index repository"));
        printf("%s\n", _TEXT("Options:"));
        shell::help();
        printf("\n%s\n", _TEXT("Report bugs to dyfet@gnu.org"));
        PROGRAM_EXIT(0);
    }

    archive = argv[--argc];

    if(!fsys::is_dir(archive))
        shell::errexit(1, "*** cape-archive: %s: %s\n",
            archive, _TEXT("not a directory"));

    plock lock(strdup(str(archive) + "/.lock"), 1000);

    unsigned index = 0;
    while(index < (unsigned)argc) {
        char *pkg = argv[index++];

        if(!fsys::is_file(pkg)) {
            fprintf(stderr, "*** cape-archive: %s: %s\n",
                pkg, _TEXT("no such file"));
            continue;
        }

        // silently skip .orig.tar.gz and .debian.tar.gz...
        if(strstr(pkg, ".orig.tar.gz"))
            continue;

        if(strstr(pkg, ".debian.tar.gz"))
            continue;

        char *base = strrchr(pkg, '/');
        const char *dir = ".";
        if(base) {
            dir = pkg;
            *(base++) = 0;
        }
        else
            base = pkg;
        const char *ext = strrchr(base, '.');
        char *tar = strchr(base, '.');
        char *version = strchr(base, '-');
        while(version && !isdigit(version[1]))
            version = strchr(++version, '-');
        while(tar && !eq(tar, ".tar.gz"))
            tar = strchr(++tar, '.');
        if(tar && version) {
            *(tar++) = 0;
            *(version++) = 0;
            tar_update(dir, base, version, tar, archive);
        }
        else if(eq(ext, ".rpm"))
            rpm_update(dir, base, archive);
        else if(eq(ext, ".dsc") || eq(ext, ".changes"))
            deb_update(dir, base, archive);
        else
            fprintf(stderr, "*** cape-archive: %s: %s\n",
                _STR(str(dir) + "/" + (const char *)base), _TEXT("unrecognized pkg file"));
    }

    if(is(ignore))
        PROGRAM_EXIT(0);

    unsigned deb = 0, rpm = 0, dsc = 0, gz = 0;
    dir_t dir(archive);
    while(is(dir) && dir.read(filename, sizeof(filename)) > 0) {
        ext = strrchr(filename, '.');
        if(eq(ext, ".deb"))
            ++deb;
        else if(eq(ext, ".dsc"))
            ++dsc;
        else if(eq(ext, ".rpm"))
            ++rpm;
        else if(eq(ext, ".gz"))
            ++gz;
    }
    dir.close();

    // pure tarball archives do not have indexes...
    if(!deb && !rpm && !dsc && gz)
        PROGRAM_EXIT(0);

    if(!deb && !rpm && !dsc) {
        fprintf(stderr, "*** cape-archive: %s: %s\n",
            archive, _TEXT("no packages"));
        PROGRAM_EXIT(0);
    }

    if(chdir(archive))
        shell::errexit(10, "*** cape-archive: %s: %s",
            archive, _TEXT("invalid directory"));
    fsys::prefix(prefix, sizeof(prefix));
    printf("indexing %s...\n", prefix);

    if(rpm)
        rpm_index(prefix);

    remove(str(prefix) + "/repodata/repomd.xml.asc");
    if(is(sign) && rpm)
        rpm_release(prefix);

    char *subdir = strrchr(prefix, '/');
    if(subdir)
        *(subdir++) = 0;

    if(dsc)
        dsc_index(prefix, subdir);

    if(deb)
        deb_index(prefix, subdir);

    if(is(sign) && (deb | dsc))
        deb_release(prefix, subdir);
    else {
        String release = str(subdir) + "/Release";
        remove(release);
        release += ".gpg";
        remove(release);
    }
    PROGRAM_EXIT(0);
}

plock::plock(char *lockfile, timeout_t timer)
{
    pid_t locker;
    char buf[32];
    path = NULL;
    int mask = umask(002);
    unsigned long count = 0;

    for(;;) {
        if(count++ == 1) {
            printf("waiting...");
            fflush(stdout);
        }

        fd_t fd = open(lockfile, O_WRONLY | O_CREAT | O_EXCL, 0660);
        if(fd < 0) {
            fd = open(lockfile, O_RDONLY);
            if(read(fd, buf, sizeof(buf) > 0)) {
                // see if locked by another...
                locker = atol(buf);
                if(!kill(locker, 0)) {
                    close(fd);
                    Thread::sleep(timer);
                    continue;
                }
                remove(lockfile);
                continue;
            }
            else {
                time_t now;
                struct stat ino;
                stat(lockfile, &ino);
                time(&now);
                // if not set with pid for 5 seconds...stale
                if(ino.st_mtime < now - 5)
                    remove(lockfile);
                Thread::sleep(timer);
                continue;
            }
        }
        if(count > 1)
            printf("\n");
        path = lockfile;
        pid = getpid();
        snprintf(buf, sizeof(buf), "%ld", (long)pid);
        if(write(fd, buf, strlen(buf) + 1) < 1)
            shell::errexit(-1, "*** cape-archive: %s",
                _TEXT("lock failed write"));
        close(fd);
        break;
    }
    umask(mask);
}

plock::~plock()
{
    release();
}

void plock::release(void)
{
    if(path && pid == getpid()) {
        fsys::erase(path);
        free(path);
        path = NULL;
    }
}

