// Copyright (C) 2012 David Sugar, Tycho Softworks.
//
// This file is part of  cape-devtools.
//
// Cape-devtools is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Cape-devtools is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Cape-devtools.    If not, see <http://www.gnu.org/licenses/>.

#include <cape-devtools-config.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/wait.h>
#include <sys/utsname.h>
#include <sys/fcntl.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <ucommon/ucommon.h>

using namespace UCOMMON_NAMESPACE;

static shell::flagopt helpflag('h',"--help",    _TEXT("display this list"));
static shell::flagopt althelp('?', NULL, NULL);
static shell::stringopt profile('p', "--profile", _TEXT("project profile"), "config", "default");
static shell::flagopt ppa('P', "--ppa", _TEXT("ppa mode"));
static shell::numericopt buildrev('r', "--revision", _TEXT("build revision"), ">1", 1);
static shell::flagopt sign('s', "--sign", _TEXT("sign index"));
static shell::flagopt verbose('v', "--verbose", _TEXT("verbose"));

static const char *out = NULL;
static const char *realname, *email, *signid = NULL;
static struct tm *dt;
static bool redhatos = false, debianos = false;
static bool merge = true;

static const char *months[12] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

static const char *days[7] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};

static void erasefiles(const char *path, const char *prefix, const char *suffix)
{
    char buffer[65];
    dir_t dir(path);
    size_t plen = strlen(prefix), slen = strlen(suffix), len;
    while(is(dir) && dir.read(buffer, sizeof(buffer)) > 0) {
        len = strlen(buffer);
        if(len <= slen)
            continue;
        if(!eq(prefix, buffer, plen))
            continue;
        if(!eq(suffix, buffer + len - slen))
            continue;
        if(is(verbose))
            printf("removing %s\n", buffer);
        fsys::erase(str(path) + "/" + (const char *)buffer);
    }
}

static void copyfile(String& from, String& to)
{
    char buffer[4096];

    fd_t source, target;
    fsys::fileinfo_t finfo, tinfo;
    String tmp = str(to) + ".tmp.";
    if(fsys::info(*from, &finfo))
        return;

    // if file unchanged, no need to copy
    if(!fsys::info(*to, &tinfo)) {
        if(finfo.st_mtime < tinfo.st_mtime && finfo.st_size == tinfo.st_size)
            return;
    }

    source = open(*from, O_RDONLY);
    if(source < 0)
        return;

    remove(*tmp);

    // if another session already doing it, ignore...
    target = open(*tmp, O_CREAT | O_EXCL | O_WRONLY, finfo.st_mode & 0777);
    if(target < 0) {
        close(source);
        return;
    }

    if(is(verbose))
        printf("copying %s\n", *from);

    for(;;) {
        ssize_t len = read(source, buffer, sizeof(buffer));
        if(len < 0) {
            close(source);
            close(target);
            remove(*tmp);
            fprintf(stderr, "*** deb-backport: %s: %s\n",
                *from, _TEXT("unable to copy"));
        }
        if(len > 0) {
            if(write(target, buffer, len) < len)
                shell::errexit(10, "*** deb-backport: %s",
                    _TEXT("failed to copy"));
        }
        if((size_t)len < sizeof(buffer))
            break;
    }
    close(source);
    close(target);
    rename(*tmp, *to);
}

static void deb_publish(const char *dir, const char *base, char **dists)
{
    String dscfile = base;
    String fname;
    String from = str(dir) + "/" + base;
    String to = str(out) + "/" + base;
    file_t changes(*from, "r");
    bool files = false;
    char *fn = NULL, *fe, *fext;
    char *version = (char *)"", *revision = (char *)"-0";
    char buffer[256];
    const char *ext = strrchr(base, '.');
    char *ep = (char *)strchr(base, '_');
    char *lp = (char *)strrchr(base, '_');
    ep = strchr((char *)base, '_');
    if(ep)
        *ep = 0;

    while(changes.getline(buffer, sizeof(buffer))) {
        if(eq(buffer, "Version:", 8)) {
            version = buffer + 8;
            while(isspace(*version))
                ++version;
            version = strdup(version);
            revision = strchr(version, '-');
            if(revision)
                *(revision++) = 0;
            continue;
        }
        if(eq(buffer, "Files:", 6)) {
            files = true;
            continue;
        }
        if(buffer[0] != ' ' || !files) {
            files = false;
            continue;
        }
        fn = strchr(buffer + 2, ' ');
        while(NULL != (fe = strchr(++fn, ' '))) {
            fn = fe;
        }
        fname = str(fn);
        fe = strchr(fn, '_');
        if(!fe)
            continue;
        *(fe++) = 0;
        fext = strstr(fe, ".orig.");
        if(!fext)
            fext = strchr(fe, '_');
        if(!fext)
            fext = fe + strlen(version) + strlen(revision) + 1;

        erasefiles(out, str(fn) + "_", fext);
        String ffrom = str(dir) + "/" + fname;
        String fto = str(out) + "/" + fname;
        copyfile(ffrom, fto);
    }

    if(ep != lp)
        ext = lp;
    erasefiles(out, str(base) + "_", ext);
    copyfile(from, to);

    if(chdir(out))
        shell::errexit(3, "*** deb-backport: %s: %s",
            out, _TEXT("cannot access directory"));

    char *argv[32];
    unsigned index = 0;

    String src = str(fn) + "-" + (const char *)version;

    argv[index++] = (char *)"dpkg-source";
    argv[index++] = (char *)"-x";
    argv[index++] = (char *)*dscfile;
    argv[index++] = NULL;

    int status = fsys::exec(*argv, argv);
    if(status)
        exit(status);

    String newchanges = str(src) + "/debian/changelog";
    String oldchanges = str(fn) + ".changelog";
    copyfile(newchanges, oldchanges);

    char change[256];
    file_t chfile(oldchanges, "r");
    chfile.getline(change, sizeof(change));
    const char *chmode = strrchr(change, ';');
    const char *dist;

    if(!chmode)
        chmode = ";";

    tzset();
    if(daylight)
        timezone -= 3600l;
    timezone /= 60l;
    int tzh = timezone / 60l;
    int tzm = timezone % 60l;
    const char *tzsign = "-";

    if(tzh >= 12) {
        tzh = 24 - tzh;
        tzsign = "+";
    }

    const char *incsrc = "-sa";

    if(is(buildrev))
        incsrc = "-sd";

    String gpg;
    if(signid)
        gpg = str("gpg -u ") + signid;
    else
        gpg = str("gpg");

    while(NULL != (dist = *(dists++))) {
        chfile.rewind();
        String distver;
        if(is(ppa))
            distver = str(version) + "-" + str(revision) + "~" + dist + "1~ppa" + str(*buildrev);
        else
            distver = str(version) + "-" + str(revision) + "~bp" + str(*buildrev);
        if(is(verbose))
            printf("packaging for %s\n", *distver);

        FILE *fp = fopen(*newchanges, "w");

        if(!fp) {
            shell::errexit(5, "*** deb-backport: %s: %s",
                *newchanges, _TEXT("cannot create"));
        }

        printf("newfile %s\n", *newchanges);

        fprintf(fp, "%s (%s) %s%s\n\n", fn, *distver, dist, chmode);

        char buffer[256];
        if(merge) {
            while(chfile.getline(buffer, sizeof(buffer)) && buffer[0])
                ;

            while(chfile.getline(buffer, sizeof(buffer)) && buffer[0])
                fprintf(fp, "%s\n", buffer);
        }
        else
            fprintf(fp, "  * No change backport for publication\n");


        fprintf(fp, "\n -- %s <%s>  %s, %02d %s %d %02d:%02d:%02d %s%02d%02d\n\n",
            realname, email, days[dt->tm_wday], dt->tm_mday,
            months[dt->tm_mon], dt->tm_year + 1900,
            dt->tm_hour, dt->tm_min, dt->tm_sec, tzsign, tzh, tzm);

        while(!merge && chfile.getline(buffer, sizeof(buffer)))
            fprintf(fp, "%s\n", buffer);

        fclose(fp);

        char *cmdv[32];
        unsigned index = 0;

        cmdv[index++] = (char *)"dpkg-source";
        cmdv[index++] = (char *)"-b";
        cmdv[index++] = (char *)*src;
        cmdv[index++] = NULL;

        int status = fsys::exec(*cmdv, cmdv);
        if(status)
            exit(status);

        index = 0;

        cmdv[index++] = (char *)"dpkg-genchanges";
        cmdv[index++] = (char *)"-S";
        cmdv[index++] = (char *)incsrc;
        cmdv[index++] = NULL;

        String changefile = str(fn) + "_" + distver + "_source.changes";
        String dscfile = str(fn) + "_" + distver + ".dsc";

        pid_t pid = fork();
        if(!pid) {
            int fd = open(changefile, O_WRONLY | O_TRUNC | O_CREAT, 0644);
            if(fd < 1)
                shell::errexit(11, "*** deb-backport: %s: %s",
                    *changefile, _TEXT("cannot write"));
            dup2(fd, 1);
            if(chdir(*src))
                shell::errexit(10, "*** deb-backport: %s: %s",
                    *src, _TEXT("invalid directory"));
            execvp(*cmdv, cmdv);
            exit(-1);
        }
        shell::wait(pid);
        incsrc = "-sd";

        if(is(sign)) {
            remove(".tmp.sig");
            if(system(str(gpg) + " --clearsign --output .tmp.sig " + dscfile))
                shell::errexit(8, "*** deb-backport: %s: %s\n",
                    *dscfile, _TEXT("failed to sign"));
            if(system(str("mv -f .tmp.sig ") + dscfile))
                fprintf(stderr, "*** deb-backport: %s\n",
                    _TEXT("cannot update signature"));
        }

        if(is(sign)) {
            remove(".tmp.sig");
            if(system(str(gpg) + " --clearsign --output .tmp.sig " + changefile))
                shell::errexit(8, "*** deb-backport: %s: %s\n",
                    *changefile, _TEXT("failed to sign"));
            if(system(str("mv -f .tmp.sig ") + changefile))
                fprintf(stderr, "*** deb-backport: %s\n",
                    _TEXT("cannot update signature"));
        }
    }
}

PROGRAM_MAIN(argc, argv)
{
    shell::bind("deb-backport");
    shell args(argc, argv);
    argv = args.argv();

    char hostname[256];
    struct passwd *pwd = getpwuid(getuid());
    struct keydata *archives;
    keyfile pkgbuildrc;
    time_t now;

    if(fsys::is_file("/etc/debian_version"))
        debianos = true;

    if(fsys::is_dir("/etc/yum"))
        redhatos = true;

    time(&now);
    dt = localtime(&now);

    pkgbuildrc.load(shell::path(shell::USER_HOME) + "/.pkgbuildrc");
    archives = pkgbuildrc.get(*profile);
    gethostname(hostname, sizeof(hostname));

    if(pwd) {
        email = strdup(str(pwd->pw_name) + "@" + (const char *)hostname);
        realname = strdup(pwd->pw_gecos);
        char *ep = strchr((char *)realname, ',');
        if(ep)
            *ep = 0;
    }
    else {
        email = strdup(str(getlogin()) + "@" + (const char *)hostname);
        realname = strdup(getlogin());
    }

    if(archives && is(ppa))
        out = archives->get("ppadir");
    else if(archives)
        out = archives->get("backports");
    if(!out)
        out = "~/ppa";

    if(eq(out, "~/", 2))
        out = strdup(str(pwd->pw_dir) + ++out);

    mkdir(out, 0775);
    char *pkg = *(argv++);

    if(!pkg || is(helpflag) || is(althelp)) {
        printf("%s: deb-backport %s\n", _TEXT("Usage"),
            _TEXT("[options] .dscfile [distributions...]"));
        printf("%s\n\n", _TEXT("Create backport"));
        printf("%s\n", _TEXT("Options:"));
        shell::help();
        printf("\n%s\n", _TEXT("Report bugs to dyfet@gnu.org"));
        PROGRAM_EXIT(0);
    }

    char *base = strrchr(pkg, '/');
    const char *dir = ".";
    if(base) {
        dir = pkg;
        *(base++) = 0;
    }
    else
        base = pkg;

    if(!*argv) {
        const char *dists = archives->get("backports");
        if(dists)
            argv = args.shell::parse(dists);
    }

    const char *ext = strrchr(base, '.');
    if(eq(ext, ".dsc"))
        deb_publish(dir, base, argv);
    else
        shell::errexit(2, "*** deb-backport: %s: %s",
            pkg, _TEXT("unrecognized source package"));

    PROGRAM_EXIT(0);
}


