// Copyright (C) 2012 David Sugar, Tycho Softworks.
//
// This file is part of  cape-devtools.
//
// Cape-devtools is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Cape-devtools is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Cape-devtools.    If not, see <http://www.gnu.org/licenses/>.

#include <cape-devtools-config.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/wait.h>
#include <sys/utsname.h>
#include <sys/fcntl.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <ucommon/ucommon.h>

using namespace UCOMMON_NAMESPACE;

static shell::flagopt helpflag('h',"--help",    _TEXT("display this list"));
static shell::flagopt althelp('?', NULL, NULL);
static shell::stringopt profile('p', "--profile", _TEXT("project profile"), "config", "default");
static shell::flagopt sign('s', "--sign", _TEXT("signing mode"));
static shell::flagopt verbose('v', "--verbose", _TEXT("verbose"));
static shell::flagopt who('w', "--who", _TEXT("who am i"));

static keydata *archives = NULL;
static char prefix[512];
static keyfile pkgbuildrc;
static pid_t pid;
static const char *arch = NULL;
static const char *label = NULL, *origin = NULL, *distribution = "unstable";
static const char *realname, *email, *vendor = NULL, *build_prefix = NULL;
static bool debianos = false, redhatos = false;
static int exit_code = 0;
static const char *signid = NULL;

static void copyfile(string_t from, string_t to)
{
    char buffer[4096];

    fd_t source, target;
    fsys::fileinfo_t finfo, tinfo;
    String tmp = str(to) + ".tmp.";
    if(fsys::info(*from, &finfo))
        return;

    // if file unchanged, no need to copy
    if(!fsys::info(*to, &tinfo)) {
        if(finfo.st_mtime < tinfo.st_mtime && finfo.st_size == tinfo.st_size)
            return;
    }

    source = open(*from, O_RDONLY);
    if(source < 0)
        return;

    remove(*tmp);

    // if another session already doing it, ignore...
    target = open(*tmp, O_CREAT | O_EXCL | O_WRONLY, finfo.st_mode & 0777);
    if(target < 0) {
        close(source);
        return;
    }

    if(is(verbose))
        printf("copying %s\n", *from);

    for(;;) {
        ssize_t len = read(source, buffer, sizeof(buffer));
        if(len < 0) {
            close(source);
            close(target);
            remove(*tmp);
            fprintf(stderr, "*** cape-binary: %s: %s\n",
                *from, _TEXT("unable to copy"));
        }
        if(len > 0) {
            if(write(target, buffer, len) < len)
                shell::errexit(10, "*** cape-binary: %s",
                    _TEXT("failed to copy"));
        }
        if((size_t)len < sizeof(buffer))
            break;
    }
    close(source);
    close(target);
    rename(*tmp, *to);
}

static void erasefiles(const char *path, const char *prefix, const char *suffix)
{
    char buffer[65];
    dir_t dir(path);
    size_t plen = strlen(prefix), slen = strlen(suffix), len;
    while(is(dir) && dir.read(buffer, sizeof(buffer)) > 0) {
        len = strlen(buffer);
        if(len <= slen)
            continue;
        if(!eq(prefix, buffer, plen))
            continue;
        if(!eq(suffix, buffer + len - slen))
            continue;
        if(is(verbose))
            printf("removing %s\n", buffer);
        fsys::erase(str(path) + "/" + (const char *)buffer);
    }
}

static void cleanexit(void)
{
    char *argv[4];
    if(pid != getpid() || !build_prefix)
        return;

    pid = 0;
    argv[0] = (char *)"rm";
    argv[1] = (char *)"-rf";
    argv[2] = (char *)build_prefix;
    argv[3] = NULL;

    shell::release(exit_code);
    fsys::exec(*argv, argv);
}

static void deb_arch(void)
{
    char buffer[80];
    FILE *fp = popen("dpkg --print-architecture", "r");
    if(!fp || !fgets(buffer, sizeof(buffer), fp)) {
        if(fp)
            pclose(fp);
        shell::errexit(20, "*** cape-binary: %s",
            _TEXT("cannot use dpkg"));
    }
    pclose(fp);
    char *cp = strpbrk(buffer, "\r\n");
    if(cp)
        *cp = 0;
    arch = strdup(buffer);
}

static void rpm_arch(void)
{
    char buffer[80];
    FILE *fp = popen("rpm --eval %_arch", "r");
    if(!fp || !fgets(buffer, sizeof(buffer), fp)) {
        if(fp)
            pclose(fp);
        shell::errexit(20, "*** cape-binary: %s",
            _TEXT("cannot use rpm"));
    }
    pclose(fp);
    char *cp = strpbrk(buffer, "\r\n");
    if(cp)
        *cp = 0;
    arch = strdup(buffer);
}

static void deb_sign(const char *path)
{
    string_t gpg;

    if(signid)
        gpg = str("gpg -u ") + signid;
    else
        gpg = str("gpg");

    remove(".tmp.sig");
    if(system(str(gpg) + " --clearsign --output .tmp.sig " + path)) {
        shell::errexit(8, "*** cape-source: %s: %s\n",
            path, _TEXT("failed to sign"));
    }
    if(system(str("mv -f .tmp.sig ") + path)) {
        fprintf(stderr, "*** cape-source: %s: %s\n",
            path, _TEXT("cannot update signature"));
    }
}

static void rpm_sign(char *path)
{
    char *argv[16];
    unsigned index = 0;

    string_t packager = str("_packager ") + (const char *)realname + " <" + (const char *)email + ">";

    argv[index++] = (char *)"rpmbuild";
    argv[index++] = (char *)"--define";
    argv[index++] = (char *)*packager;

    if(vendor && *vendor) {
        string_t v = str("_vendor ") + vendor;
        argv[index++] = (char *)"--define";
        argv[index++] = (char *)*v;
    }

    if(signid) {
        string_t signer = str("_gpg_name ") + (const char *)realname + " <" + (const char *)signid + ">";
        argv[index++] = (char *)"--define";
        argv[index++] = (char *)*signer;
    }

    if(is(verbose))
        argv[index++] = (char *)"-v";

    argv[index++] = (char *)"--sign";
    argv[index++] = (char *)path;
    argv[index] = NULL;

    exit_code = fsys::exec(*argv, argv);
    if(exit_code)
        shell::errexit(10, "*** cape-binary: %s: %s",
            path, _TEXT("failed to sign"));
}

static void deb_rebuild(char *path)
{
    char *argv[8];
    string_t source = path;
    char *dscfile = strrchr(path, '/');
    if(dscfile)
        ++dscfile;
    else
        dscfile = path;
    char *version = strchr(dscfile, '_');
    if(version)
        *(version++) = 0;
    else
        shell::errexit(11, "*** cape-binary: %s: %s",
            *source, _TEXT("unknown source package"));
    char *suffix = strchr(version, '-');
    if(suffix)
        *suffix = 0;

    string_t dscprefix = str(dscfile) + "_";

    erasefiles(prefix, dscprefix, str("_") + arch + ".deb");
    erasefiles(prefix, dscprefix, str("_") + arch + ".changes");
    erasefiles(prefix, dscprefix, "_all.deb");

    string_t build = str(build_prefix) + "/" + (const char *)dscfile + "-" + (const char *)version;

    argv[0] = (char *)"dpkg-source";
    argv[1] = (char *)"-x";
    argv[2] = (char *)*source;
    argv[3] = (char *)*build;
    argv[4] = NULL;
    exit_code = fsys::exec(*argv, argv);
    if(exit_code)
        shell::errexit(exit_code, "*** cape-binary: %s: %s",
            *source, _TEXT("failed to unpack source"));

    if(chdir(*build))
        shell::errexit(2, "*** cape-binary: %s: %s",
            *build, _TEXT("cannot access directory"));
    argv[0] = (char *)"dpkg-buildpackage";
    argv[1] = (char *)"-b";
    argv[2] = (char *)"-uc";
    argv[3] = NULL;

    exit_code = fsys::exec(*argv, argv);
    if(exit_code)
        shell::errexit(exit_code, "*** cape-binary: %s: %s",
            *source, _TEXT("failed to build source"));

    if(chdir(".."))
        shell::errexit(2, "*** cape-binary: %s: %s",
            "..", _TEXT("cannot access parent directory"));
    char buffer[256];
    size_t len = strlen(dscprefix);
    dir_t dir(".");
    if(!is(dir))
        ::exit(-1);

    file_t changes;
    string_t chfile;
    while(dir.read(buffer, sizeof(buffer)) > 0) {
        const char *ext = strrchr(buffer, '.');
        if(eq(buffer, dscprefix, len) && eq(ext, ".changes")) {
            changes.open(buffer, "r");
            chfile = buffer;
            break;
        }
    }
    dir.close();

    if(!is(changes))
        shell::errexit(12, "*** cape-binary: %s: %s",
            *source, _TEXT("cannot find changes"));

    bool files = false;
    const char *fn, *fe;
    while(changes.getline(buffer, sizeof(buffer)) > 0) {
        if(eq(buffer, "Files:", 6)) {
            files = true;
            continue;
        }
        if(buffer[0] != ' ' || !files) {
            files = false;
            continue;
        }
        fn = strchr(buffer + 2, ' ');
        while(NULL != (fe = strchr(++fn, ' '))) {
            fn = fe;
        }
        string_t fname = fn;
        copyfile(fname, str(prefix) + "/" + fn);
    }
    copyfile(chfile, str(prefix) + "/" + chfile);
    changes.close();
}

static void rpm_rebuild(const char *path)
{
    char *argv[32];
    unsigned index = 0;
    char *tarbase = strdup(path);
    char *ep = tarbase;

    while((ep = strchr(ep, '-')) != NULL)
    {
        if(*ep == '-' && isdigit(ep[1])) {
            *ep = 0;
            break;
        }
        ++ep;
    }

    erasefiles(prefix, str(tarbase) + "-", str(".") + arch + ".rpm");

    string_t sourcedir = str("_sourcedir ") + build_prefix;
    string_t builddir = str("_buildroot ") + build_prefix;
    string_t specdir = str("_builddir ") + build_prefix;
    string_t rpmdir = str("_rpmdir ") + build_prefix;
    string_t packager = str("_packager ") + (const char *)realname + " <" + (const char *)email + ">";

    argv[index++] = (char *)"rpmbuild";
    argv[index++] = (char *)"--define";
    argv[index++] = (char *)*packager;
    argv[index++] = (char *)"--define";
    argv[index++] = (char *)*sourcedir;
    argv[index++] = (char *)"--define";
    argv[index++] = (char *)*rpmdir;
    argv[index++] = (char *)"--define";
    argv[index++] = (char *)*builddir;
    argv[index++] = (char *)"--define";
    argv[index++] = (char *)*specdir;

    if(vendor && *vendor) {
        String v = str("_vendor ") + vendor;
        argv[index++] = (char *)"--define";
        argv[index++] = (char *)*v;
    }

    if(is(verbose))
        argv[index++] = (char *)"-v";

    argv[index++] = (char *)"--clean";
    argv[index++] = (char *)"--rebuild";
    argv[index++] = (char *)path;
    argv[index] = NULL;

    exit_code = fsys::exec(*argv, argv);
    if(exit_code)
        shell::errexit(10, "*** cape-binary: %s: %s",
            path, _TEXT("failed to build"));
}

PROGRAM_MAIN(argc, argv)
{
    shell::bind("cape-binary");
    shell args(argc, argv);
    argv = args.argv();
    char filename[65];
    char hostname[256];

    struct passwd *pwd = getpwuid(getuid());

    if(fsys::is_file("/etc/debian_version"))
        debianos = true;

    if(fsys::is_dir("/etc/yum"))
        redhatos = true;

    pid = getpid();

    pkgbuildrc.load(shell::path(shell::USER_HOME) + "/.pkgbuildrc");
    archives = pkgbuildrc.get(*profile);
    gethostname(hostname, sizeof(hostname));

    if(pwd) {
        email = strdup(str(pwd->pw_name) + "@" + (const char *)hostname);
        realname = strdup(pwd->pw_gecos);
        char *ep = strchr((char *)realname, ',');
        if(ep)
            *ep = 0;
    }
    else {
        email = strdup(str(getlogin()) + "@" + (const char *)hostname);
        realname = strdup(getlogin());
    }

    if(fsys::is_file("/etc/redhat_release"))
        rpm_arch();
    else
        deb_arch();

    if(archives) {
        label = archives->get("label");
        origin = archives->get("origin");
        vendor = archives->get("vendor");
        distribution = archives->get("distribution");
        const char *em = archives->get("email");
        if(em)
            email = em;

        signid = archives->get("key");
        if(!signid)
            signid = em;
        if(em)
            email = em;

        if(!vendor || !*vendor)
            vendor = label;

        if(!vendor || !*vendor)
            vendor = distribution;
    }

    setenv("DEBEMAIL", email, 1);
    setenv("DEBFULLNAME", realname, 1);

    fsys::prefix(prefix, sizeof(prefix));

    if(is(helpflag) || is(althelp)) {
        printf("%s: cape-binary %s\n", _TEXT("Usage"),
            _TEXT("[options] sourcepkg"));
        printf("%s\n\n", _TEXT("Create binary packages"));
        printf("%s\n", _TEXT("Options:"));
        shell::help();
        printf("\n%s\n", _TEXT("Report bugs to dyfet@gnu.org"));
        PROGRAM_EXIT(0);
    }

    if(!*argv) {
        shell::errexit(1, "*** cape-binary: %s\n",
            _TEXT("no source package to build"));
    }


    build_prefix = strdup(str("/tmp/buildd-") + str((long)getpid()));
    pid = getpid();
    atexit(cleanexit);

    while(*argv) {
        mkdir(build_prefix, 0755);
        if(chdir(prefix))
            shell::errexit(2, "*** cape-binary: %s: %s",
                prefix, _TEXT("cannot access directory"));
        const char *ext = strrchr(*argv, '.');
        const char *sub = strrchr(*argv, '_');
        if(eq(ext, ".dsc") && debianos)
            deb_rebuild(*argv);
        else if(is(sign) && eq(ext, ".changes") && sub && !eq(sub, "_source.changes"))
            deb_sign(*argv);
        else if((ext > *argv + 4) && eq(ext - 4, ".src.rpm") && redhatos)
            rpm_rebuild(*argv);
        else if(eq(ext, ".rpm") && redhatos && is(sign))
            rpm_sign(*argv);
        else if(is(sign))
            shell::errexit(2, "*** cape-binary: %s: %s",
                filename, _TEXT("invalid binary package to sign"));
        else
            shell::errexit(2, "*** cape-binary: %s: %s",
                filename, _TEXT("unknown or invalid source package"));
        ++argv;
    }

    PROGRAM_EXIT(0);
}


