// Copyright (C) 2012 David Sugar, Tycho Softworks.
//
// This file is part of  cape-devtools.
//
// Cape-devtools is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Cape-devtools is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Cape-devtools.    If not, see <http://www.gnu.org/licenses/>.

#include <cape-devtools-config.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/wait.h>
#include <sys/utsname.h>
#include <sys/fcntl.h>
#include <unistd.h>
#include <pwd.h>
#include <ctype.h>
#include <ucommon/ucommon.h>

using namespace UCOMMON_NAMESPACE;

static shell::flagopt helpflag('h',"--help",    _TEXT("display this list"));
static shell::flagopt althelp('?', NULL, NULL);
static shell::flagopt amend('a', "--amend", _TEXT("modify existing"));
static shell::stringopt profile('p', "--profile", _TEXT("project profile"), "config", "default");
static shell::flagopt revised('r', "--revise", _TEXT("debian revision"));
static shell::stringopt urgency('u', "--urgency", _TEXT("priority of change"), "level", "low");
static shell::flagopt verinfo('V', "--version-info", _TEXT("print spec version info"));
static shell::flagopt relinfo('R', "--release-info", _TEXT("print spec release info"));

static const char *changes;
static const char *newfile;
static const char *version = NULL;
static const char *editor = "vim";
static const char *distribution = "UNRELEASED";
static const char *pkgname = "???";
static const char *realname = NULL, *email = NULL;
static struct tm *dt;
static keyfile pkgbuildrc;

static const char *months[12] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

static const char *days[7] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};

static const char *next_package(const char *version, const char *prior)
{
    long value = 0;
    char *head = strdup(prior);
    char *tail = NULL;
    char *save = head;

    while(*head) {
        if(isdigit(*head)) {
            if(!tail)
                tail = head;
            value *= 10;
            value += (*head - '0');
        }
        else
            tail = NULL;
        ++head;
    }
    ++value;
    if(tail)
        *tail = 0;
    return strdup(str(version) + "-" + (const char *)save + str(value));
}

static const char *next_revision(const char *version, const char *prior)
{
    if(!is(revised))
        return next_package(version, prior);

    long value = 0;
    while(*prior && isdigit(*prior)) {
        value *= 10;
        value += (*prior - '0');
        ++prior;
    }
    ++value;
    return strdup(str(version) + "-" + str(value) + prior);
}

static FILE *rpm_header(const char *reason)
{
    FILE *fp = fopen(newfile, "w");

    if(!fp) {
        shell::errexit(5, "*** cape-changes: %s: %s",
            newfile, _TEXT("cannot create"));
    }

    fprintf(fp, "* %s %s %02d %04d %s <%s> - %s\n"
                "- %s\n\n",
        days[dt->tm_wday], months[dt->tm_mon], dt->tm_mday,
        dt->tm_year + 1900, realname, email, version, reason);
    return fp;
}

static FILE *deb_header(const char *reason)
{
    FILE *fp = fopen(newfile, "w");

    tzset();
    if(daylight)
        timezone -= 3600l;

    timezone /= 60l;
    int tzh = timezone / 60l;
    int tzm = timezone % 60l;
    const char *tzsign = "-";

    if(!fp) {
        shell::errexit(5, "*** cape-changes: %s: %s",
            newfile, _TEXT("cannot create"));
    }
    fprintf(fp, "%s (%s) %s; urgency=%s\n\n"
                "  * %s\n\n"
                " -- %s <%s>  %s, %02d %s %d %02d:%02d:%02d %s%02d%02d\n\n",
        pkgname, version, distribution, *urgency, reason,
        realname, email, days[dt->tm_wday], dt->tm_mday,
        months[dt->tm_mon], dt->tm_year + 1900,
        dt->tm_hour, dt->tm_min, dt->tm_sec, tzsign, tzh, tzm);
    return fp;
}

static int edit(const char *path, unsigned line)
{
    char *argv[8];
    unsigned index = 0;

    argv[index++] = (char *)editor;
    if(line) {
        char buf[16];
        snprintf(buf, sizeof(buf), "+%d", line);
        argv[index++] = strdup(buf);
    }
    argv[index++] = (char *)path;
    argv[index++] = NULL;

    return fsys::exec(*argv, argv);
}

static FILE *amend_file(void)
{
    FILE *fp = fopen(newfile, "w");

    if(!fp) {
        shell::errexit(5, "*** cape-changes: %s: %s",
            newfile, _TEXT("cannot create"));
    }

    return fp;
}

static void deb_create(void)
{
    if(!version)
        version = "0.1.0-1";

    fsys::fileinfo_t before, after;
    FILE *fp = deb_header("Initial release");
    fclose(fp);
    fsys::info(newfile, &before);
    int excode = edit(newfile, 3);
    fsys::info(newfile, &after);
    if(excode || before.st_mtime == after.st_mtime)
        remove(newfile);
    else
        rename(newfile, changes);
    ::exit(excode);
}

static void rpm_changes(void)
{
    if(!fsys::is_file(changes))
        shell::errexit(2, "*** cape-changes: %s: %s",
            changes, _TEXT("cannot create spec"));

    file_t chfile(changes, "r");
    char buffer[256];
    FILE *fp = NULL;
    fsys::fileinfo_t before, after;

    if(!is(chfile)) {
        shell::errexit(3, "*** cape-changes: %s: %s",
            changes, _TEXT("cannot access"));
    }

    while(chfile.getline(buffer, sizeof(buffer))) {
        size_t len = strlen(buffer);
        while(len && isspace(buffer[--len]))
            buffer[len] = 0;

        if(eq(buffer, "%changelog")) {
            while(chfile.getline(buffer, sizeof(buffer))) {
                len = strlen(buffer);
                while(len && isspace(buffer[--len]))
                    buffer[len] = 0;
                if(buffer[0])
                    break;
            }
            if(!buffer[0])
                break;

            if(!is(amend) && !version) {
                char *spc = strrchr(buffer, ' ');
                char *ep = strchr(spc, '-');
                if(ep) {
                    *(ep++) = 0;
                    version = next_revision(++spc, ep);
                    *(--ep) = '-';
                }
                else
                    version = strdup(++spc);
            }

            if(!is(amend) && !strchr(version, '-'))
                    version = strdup(str(version) + "-1");

            if(!is(amend))
                fp = rpm_header("Updated...");
        }
        if(*buffer == '%' && fp)
            break;
        if(fp)
            fprintf(fp, "%s\n", buffer);
    }
    chfile.rewind();
    if(!fp) {
        if(!version)
            version = "0.1.0-1";
        if(!strchr(version, '-'))
            version = strdup(str(version) + "-1");
        fp = rpm_header("Initial release");
    }

    fclose(fp);

    fsys::info(newfile, &before);
    int excode = edit(newfile, 2);
    fsys::info(newfile, &after);
    if(excode || (before.st_mtime == after.st_mtime)) {
        remove(newfile);
        exit(excode);
    }

    char *tmpfile = strdup(str(changes) + ".tmp");
    FILE *tmp = fopen(tmpfile, "w");
    file_t nf(newfile, "r");
    if(!is(nf) || !tmp) {
        remove(newfile);
        shell::errexit(4, "*** cape-changes: %s",
            _TEXT("cannot revise specfile"));
    }

    while(chfile.getline(buffer, sizeof(buffer))) {
        size_t len = strlen(buffer);
        while(len && isspace(buffer[--len]))
            buffer[len] = 0;

        if(eq(buffer, "%changelog"))
            break;

        fprintf(tmp, "%s\n", buffer);
    }

    fprintf(tmp, "%%changelog\n");

    while(nf.getline(buffer, sizeof(buffer))) {
        size_t len = strlen(buffer);
        while(len && isspace(buffer[--len]))
            buffer[len] = 0;

        fprintf(tmp, "%s\n", buffer);
    }
    nf.close();

    while(chfile.getline(buffer, sizeof(buffer))) {
        if(buffer[0] == '%')
            break;
        buffer[0] = 0;
    }

    if(buffer[0])
        fprintf(tmp, "%s\n", buffer);

    while(chfile.getline(buffer, sizeof(buffer))) {
        size_t len = strlen(buffer);
        while(len && isspace(buffer[--len]))
            buffer[len] = 0;

        if(eq(buffer, "%changelog"))
            break;

        fprintf(tmp, "%s\n", buffer);
    }
    fclose(tmp);
    chfile.close();
    rename(tmpfile, changes);
    exit(0);
}

static void deb_changes(void)
{
    if(!is(relinfo) && !is(verinfo) && !fsys::is_file(changes)) {
        deb_create();
        return;
    }

    char buffer[256];
    file_t chfile(changes, "r");
    if(!is(chfile)) {
        shell::errexit(3, "*** cape-changes: %s: %s",
            changes, _TEXT("cannot access"));
    }

    chfile.getline(buffer, sizeof(buffer));

    char *revision;
    char *ver = strchr(buffer, '(');
    char *ep = strchr(buffer, ' ');
    if(ep)
        *ep = 0;
    pkgname = strdup(buffer);

    if(ver) {
        ++ver;
        ep = strchr(ver, ')');
        if(ep)
            *ep = 0;
    }
    if(ver)
        ep = strchr(ver, '-');
    if(ep) {
        *ep = 0;
        revision = strdup(++ep);
    }
    else
        revision = NULL;

    if(is(verinfo)) {
        ::printf("%s\n", ver);
        ::exit(0);
    }

    if(is(relinfo)) {
        if(revision)
            ::printf("%s-%s\n", ver, revision);
        else
            ::printf("%s\n", ver);
        ::exit(0);
    }

    if(version && !strchr(version, '-'))
        version = strdup(str(version) + "-1");

    if(!version && revision)
        version = next_revision(ver, revision);
    else if(!version)
        version = strdup(str(ver) + "-1");

    chfile.rewind();

    fsys::fileinfo_t before, after;
    FILE *fp;

    if(is(amend))
        fp = amend_file();
    else
        fp = deb_header("Updated...");

    while(chfile.getline(buffer, sizeof(buffer)))
        fprintf(fp, "%s\n", buffer);

    chfile.close();

    fclose(fp);
    fsys::info(newfile, &before);
    int excode = edit(newfile, 3);
    fsys::info(newfile, &after);
    if(excode || before.st_mtime == after.st_mtime)
        remove(newfile);
    else
        rename(newfile, changes);
    ::exit(excode);
}

static void tar_changes(void)
{
    if(!fsys::is_file(changes)) {
//      tar_create();
        return;
    }
}

PROGRAM_MAIN(argc, argv)
{
    shell::bind("cape-changes");
    shell args(argc, argv);
    argv = args.argv();
    char hostname[256];
    time_t now;

    time(&now);
    dt = localtime(&now);

    struct passwd *pwd = getpwuid(getuid());

    pkgbuildrc.load(shell::path(shell::USER_HOME) + "/.pkgbuildrc");
    keydata *archives = pkgbuildrc.get(*profile);
    gethostname(hostname, sizeof(hostname));

    if(pwd) {
        email = strdup(str(pwd->pw_name) + "@" + (const char *)hostname);
        realname = strdup(pwd->pw_gecos);
        char *ep = strchr((char *)realname, ',');
        if(ep)
            *ep = 0;
    }
    else {
        email = strdup(str(getlogin()) + "@" + (const char *)hostname);
        realname = strdup(getlogin());
    }

    if(archives) {
        distribution = archives->get("distribution");
        if(!distribution)
            distribution = "UNRELEASED";
        const char *em = archives->get("email");
        if(em)
            email = em;
        em = getenv("EDITOR");
        if(em)
            editor = em;
        em = archives->get("editor");
        if(em)
            editor = em;
    }


    if(is(helpflag) || is(althelp)) {
        printf("%s: cape-changes %s\n", _TEXT("Usage"),
            _TEXT("[options] changelog|specfile [version]"));
        printf("%s\n\n", _TEXT("Update changes"));
        printf("%s\n", _TEXT("Options:"));
        shell::help();
        printf("\n%s\n", _TEXT("Report bugs to dyfet@gnu.org"));
        PROGRAM_EXIT(0);
    }

    if(!*argv) {
        shell::errexit(1, "*** cape-changes: %s\n",
            _TEXT("no changelogs to update"));
    }

    changes = *(argv++);
    newfile = strdup(str(changes) + ".new");
    if(*argv)
        version = *(argv++);

    const char *fn = strrchr(changes, '/');
    if(fn)
        ++fn;
    else
        fn = changes;
    const char *ext = strchr(fn, '.');

    if(!fsys::is_file(changes) && is(amend)) {
        shell::errexit(4, "*** cape-changes: %s: %s",
            changes, _TEXT("cannot amend new file"));
    }

    if(ext && (eq(ext, ".spec") || eq(ext, ".spec.", 6)))
        rpm_changes();
    else if(eq(fn, "changelog"))
        deb_changes();
    else if(eq(fn, "ChangeLog"))
        tar_changes();
    else
        shell::errexit(3, "*** cape-changes: %s: %s",
            changes, _TEXT("unrecognized file"));
    PROGRAM_EXIT(0);
}

