// Copyright (C) 2012 David Sugar, Tycho Softworks.
//
// This file is part of  cape-devtools.
//
// Cape-devtools is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Cape-devtools is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Cape-devtools.    If not, see <http://www.gnu.org/licenses/>.

#include <cape-devtools-config.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/wait.h>
#include <sys/utsname.h>
#include <sys/fcntl.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <ucommon/ucommon.h>

#ifdef  HAVE_UNSHARE
#include <sched.h>
#endif

#ifdef  HAVE_PERSONALITY
#include <sys/personality.h>
#endif

#pragma GCC diagnostic ignored "-Wunused-result"

using namespace UCOMMON_NAMESPACE;

typedef struct
{
    char id[40];
    string_t prefix, origin, distro;
    char **argv;
}   session_t;

static shell::flagopt helpflag('h',"--help",    _TEXT("display this list"));
static shell::flagopt althelp('?', NULL, NULL);
static shell::stringopt arch('a', "--arch", _TEXT("cpu architecture"), "cpu", NULL);
static shell::flagopt binary('B', "--binary", _TEXT("binary dependent only"));
static shell::flagopt build('b', "--build", _TEXT("build packages"));
static shell::flagopt clean('c', "--clean", _TEXT("clean packages"));
static shell::stringopt xdisplay('D', "--display", _TEXT("X server display"), "socket", NULL); 
static shell::flagopt detach('d', "--detach", _TEXT("detach child process"));
static shell::flagopt force('f', "--force", _TEXT("force copy"));
static shell::stringopt home('h', "--home", _TEXT("alternate home"), "path", NULL);
static shell::numericopt jobs('j', "--jobs", _TEXT("jobs to run"), "number", 0);
static shell::flagopt localflag('l', "--local", _TEXT("local archive only"));
static shell::flagopt nodeps('n', "--nodeps", _TEXT("no build dependencies"));
static shell::stringopt newdir('p', "--prefix", _TEXT("prefix directory"));
static shell::stringopt qemu('q', "--qemu", _TEXT("qemu binary"), "path", NULL);
static shell::stringopt results('r', "--results", _TEXT("results path"), "path", NULL);
static shell::flagopt live('s', "--system", _TEXT("system mode"));
static shell::stringopt xscreen('S', "--xscreen", _TEXT("nested x server screen"), "size[:ds]", NULL);
static shell::flagopt setup(0, "--setup", _TEXT("perform setup only"));
static shell::flagopt update('u', "--update", _TEXT("update chroot"));
static shell::flagopt verbose('v', "--verbose", _TEXT("verbose logging"));
static shell::flagopt xserver('X', "--xserver", _TEXT("xserver nested mode"));

static session_t session;
static keyfile keyf;
static int exit_code = 0;
static char *mtable[64];
static unsigned mcount;
static char cwd[256];
static keydata *sys = NULL;
static struct passwd *pwd = NULL;
static struct group *grp = NULL;
static int uid, gid;
static bool root = false;
static pid_t exitpid;
static pid_t xpid = 0;

// initial default settings for live user for system chroots...
// eventually will be updated from config file...
static int liveuid = 1000;
static int livegid = 1000;
static const char *liveid = "live";
static const char *livehome = "/home/live";

#ifdef  CLONE_NEWPID
#define STACK_SIZE (1024 * 1024)
static char cstack[STACK_SIZE];
#endif

static void setenv(void)
{
    char display[256];
    char screen[256];
    char makeflags[10];
    keydata *env = NULL;

    if(*jobs > 0) {
        snprintf(makeflags, sizeof(makeflags), "-j%ld", *jobs);
        setenv("MAKEFLAGS", makeflags, 1);
    }

    if(is(xscreen)) {
        String::set(screen, sizeof(screen), *xscreen);

        const char *dp = strchr(*xscreen, ':');
        if(dp && !is(xdisplay)) {
            snprintf(display, sizeof(display), "localhost:%s", dp);
            xdisplay.set(display);
        }
        char *ep = strchr(screen, ':');
        if(ep)
            *ep = 0;

    }

    if(is(xdisplay) || is(xscreen)) {
        if(strchr(*xdisplay, ':'))
            String::set(display, sizeof(display), *xdisplay);
        else
            snprintf(display, sizeof(display), "localhost:%s", *xdisplay);
        setenv("DISPLAY", display, 1);
    } 

    if(is(xscreen)) {
        char title[256];
        
        snprintf(title, sizeof(title), "cape chroot %s", screen);
        char *args[] = {(char *)"Xephyr", (char *)"-screen", screen, (char *)"-ac", (char *)"-title", title, display, NULL};

        pid_t pid = fork();
        if(pid && pid != -1)
            xpid = pid;
        if(!pid) {
            setpgrp();
            execvp(*args, args);
            ::exit(-1); 
        }
    }

    env = keyf.get("env");
    if(!env)
        return;

    linked_pointer<keydata::keyvalue> kv = env->begin();
    while(is(kv)) {
        setenv(kv->id, kv->value, 1);
        kv.next();
    }
}

static bool sourcedir(const char *prefix)
{
    char buffer[128];
    size_t len = strlen(prefix);
    dir_t dir(".");

    if(!is(dir))
        return false;

    while(dir.read(buffer, sizeof(buffer)) > 0) {
        if(eq(buffer, prefix, len) && fsys::is_dir(buffer)) {
            if(chdir(buffer))
                shell::errexit(2, "*** cape-chroot: %s: %s",
                    buffer, _TEXT("cannot access directory"));
            return true;
        }
    }
    dir.close();
    return false;
}

static void copyfile(string_t from, string_t to)
{
    char buffer[4096];

    fd_t source, target;
    fsys::fileinfo_t finfo, tinfo;
    string_t tmp = str(to) + ".tmp.";
    if(fsys::info(*from, &finfo))
        return;

    // if file unchanged, no need to copy
    if(!fsys::info(*to, &tinfo)) {
        if(!is(force) && finfo.st_mtime < tinfo.st_mtime && finfo.st_size == tinfo.st_size)
            return;
    }

    source = open(*from, O_RDONLY);
    if(source < 0)
        return;

    if(is(force))
        remove(*tmp);

    // if another session already doing it, ignore...
    target = open(*tmp, O_CREAT | O_EXCL | O_WRONLY, finfo.st_mode & 0777);
    if(target < 0) {
        close(source);
        return;
    }

    if(is(verbose))
        printf("copying %s\n", *from);

    for(;;) {
        ssize_t len = read(source, buffer, sizeof(buffer));
        if(len < 0) {
            close(source);
            close(target);
            remove(*tmp);
            fprintf(stderr, "*** cape-chroot: %s: %s\n",
                *from, _TEXT("unable to copy"));
            exit(0);
        }
        if(len > 0)
            if(write(target, buffer, len) < len)
                shell::errexit(10, "*** cape-chroot: %s",
                    _TEXT("cannot copy"));
        if((size_t)len < sizeof(buffer))
            break;
    }
    close(source);
    close(target);
    rename(*tmp, *to);
    if(geteuid() == 0)
        if(chown(*to, 0, 0))
            printf("*** warning: %s: cannot change owner\n", *to);
}

static void dsc_copy(const char *dsc, const char *target)
{
    string_t prefix = str(cwd) + "/" + dsc;
    file_t changes(prefix, "r");
    bool files = false;
    char *fn, *fe;
    char buffer[256];

    char *dscfile = strrchr(prefix.c_mem(), '/');
    assert(dscfile != NULL);
    *(dscfile++) = 0;
    String::fix(prefix);

    while(changes.getline(buffer, sizeof(buffer)) > 0) {
        if(eq(buffer, "Files:", 6)) {
            files = true;
            continue;
        }
        if(buffer[0] != ' ' || !files) {
            files = false;
            continue;
        }
        fn = strchr(buffer + 2, ' ');
        while(NULL != (fe = strchr(++fn, ' '))) {
            fn = fe;
        }
        copyfile(str(prefix) + "/" + (const char *)fn,
            str(target) + "/" + (const char *)fn);
    }
    copyfile(str(prefix) + "/" + (const char *)dscfile,
        str(target) + "/" + (const char *)dscfile);
}

static bool deb_results(const char *prefix)
{
    file_t changes;
    bool files = false;
    char *fn, *fe;
    char buffer[256];
    const char *chfile = NULL;
    pid_t pid = fork();
    int status;

    if(pid) {
        waitpid(pid, &status, 0);
        if(WIFEXITED(status))
            exit_code = WEXITSTATUS(status);
        else
            exit_code = -1;

        if(!exit_code)
            return true;
        return false;
    }

    setgid(gid);
    setegid(gid);
    setuid(uid);
    seteuid(uid);

    size_t len = strlen(prefix);
    dir_t dir(".");
    if(!is(dir))
        ::exit(-1);

    while(dir.read(buffer, sizeof(buffer)) > 0) {
        const char *ext = strrchr(buffer, '.');
        if(eq(buffer, prefix, len) && eq(ext, ".changes")) {
            changes.open(buffer, "r");
            chfile = strdup(buffer);
            break;
        }
    }
    dir.close();

    if(!is(changes))
        ::exit(-1);

    while(changes.getline(buffer, sizeof(buffer)) > 0) {
        if(eq(buffer, "Files:", 6)) {
            files = true;
            continue;
        }
        if(buffer[0] != ' ' || !files) {
            files = false;
            continue;
        }
        fn = strchr(buffer + 2, ' ');
        while(NULL != (fe = strchr(++fn, ' '))) {
            fn = fe;
        }
        string_t fname = str(fn);
        copyfile(fname, str(*results) + "/" + fname);
    }
    string_t from = chfile;
    copyfile(from, str(*results) + "/" + chfile);
    ::exit(0);
}

static void cpu(const char *cp)
{
    if(*cp == 'i' && cp[2] == '8' && cp[3] == '6')
        arch.set("i386");
    else if(eq(cp, "x86_64"))
        arch.set("amd64");
    else
        arch.set(cp);
}


static int runuser(char **argv, const char *dir, bool cloned = false)
{
    int status;
    const char *cmd = *argv;
    const char *cp;
    static unsigned long pmode = 0;
    const char *prefix = session.prefix;
    const char *distro = session.distro;
    keydata *pk = NULL;

    if(!cloned && !is(detach)) {
        pid_t pid = fork();
        if(pid) {
            waitpid(pid, &status, 0);
            if(WIFEXITED(status))
                exit_code = WEXITSTATUS(status);
            else
                exit_code = -1;
            return exit_code;
        }
    }

    // chroot while root priv...
    if(chroot(prefix)) {
        shell::errexit(-1, "*** cape-chroot: %s: %s\n",
            prefix, _TEXT("unable to chroot; failed"));
    }

    if(root) {
        setgid(0);
        setegid(0);
        setuid(0);
        seteuid(0);
        setenv("USER", "root", 1);
        setenv("USERNAME", "root", 1);
        setenv("LOGNAME", "root", 1);
        setenv("HOME", "/root", 1);
        dir = "/root";
    }
    else if(is(live)) {
        setgid(livegid);
        setegid(livegid);
        setuid(liveuid);
        seteuid(liveuid);
        setenv("USER", liveid, 1);
        setenv("USERNAME", liveid, 1);
        setenv("LOGNAME", liveid, 1);
        setenv("HOME", livehome, 1);
        dir = livehome;
    }
    else {
        setgid(gid);
        setegid(gid);
        setuid(uid);
        seteuid(uid);
    }

    setenv();

    const char *shell, *env;
    keydata *keys = keyf.get("shell");
    if(keys)
        shell = keys->get(distro);
    else
        shell = NULL;

    if(!shell)
        shell = getenv("SHELL");
    else
        setenv("SHELL", shell, 1);

    if(!shell) {
        shell = "/bin/bash";
        setenv("SHELL", "/bin/bash", 1);
    }

    if(!chdir(dir)) {
        if(is(verbose))
            printf("changing to %s\n", dir);
    }
    else {
        fprintf(stderr, "*** cape-chroot: %s: %s\n",
            dir, _TEXT("cannot set directory"));
        if(chdir("/tmp/buildd"))
            shell::errexit(20, "*** cape-chroot: %s",
                _TEXT("cannot enter chroot for building"));
    }

    keys = keyf.get("paths");
    if(keys) {
        env = keys->get(distro);
        if(env)
            setenv("PATH", env, 1);
    }

    keys = keyf.get("prompts");
    if(keys) {
        env = keys->get(distro);
        if(env)
            setenv("PS1", env, 1);
    }

    setenv("CAPE_SESSION", session.id, 1);
    setenv("CAPE_DISTRIBUTION", distro, 1);
    setenv("CAPE_PERSONALITY", "native", 1);

    if(!is(arch)) {
        struct utsname sysinfo;
        uname(&sysinfo);
        cpu(sysinfo.machine);
    }
    else {
        cpu(*arch);
#ifdef  HAVE_PERSONALITY
        pk = keyf.get("personality");
        cp = NULL;

        // auto personality for common types...

        if(pk)
            cp = pk->get(*arch);
        else if(eq(*arch, "i386")) {
            setenv("CAPE_PERSONALITY", "linux32", 1);
            pmode = PER_LINUX32;
        }
        if(cp && eq(cp, "32")) {
            setenv("CAPE_PERSONALITY", "linux32", 1);
            pmode = PER_LINUX32;
        }
#endif
    }

    pk = keyf.get("cpu");
    cp = NULL;
    if(pk)
        cp = pk->get(*arch);
    if(cp && *cp)
        setenv("QEMU_CPU", cp, 1);

    setenv("CAPE_ARCH", *arch, 1);

    if(!*argv) {
        char *args[3];
        args[0] = (char *)shell;
        args[1] = (char *)"-l";
        args[2] = NULL;
        argv = args;
        if(is(verbose))
            printf("logging into %s with %s\n", prefix, shell);

        cmd = shell;
    }
    else {
        if(is(verbose))
            printf("executing %s\n", prefix);
        cmd = *argv;
    }
#ifdef  HAVE_PERSONALITY
    if(pmode)
        personality(pmode);
#endif
    execvp(cmd, argv);
    exit(-1);
}

static int runroot(char **argv)
{
    const char *cp;
    static unsigned long pmode = 0;
    const char *prefix = session.prefix;
    const char *distro = session.distro;
    int status;
    int pid = fork();

    if(pid) {
        waitpid(pid, &status, 0);
        if(WIFEXITED(status))
            exit_code = WEXITSTATUS(status);
        else
            exit_code = -1;
        return exit_code;
    }

    if(chroot(prefix)) {
        shell::errexit(-1, "*** cape-chroot: %s: %s\n",
            prefix, _TEXT("unable to chroot; failed"));
    }

    setgid(0);
    setegid(0);
    setuid(0);
    seteuid(0);
    setenv("USER", "root", 1);
    setenv("USERNAME", "root", 1);
    setenv("LOGNAME", "root", 1);
    setenv("HOME", "/root", 1);

    setenv("CAPE_SESSION", session.id, 1);
    setenv("CAPE_DISTRIBUTION", distro, 1);
    setenv("CAPE_PERSONALITY", "native", 1);

    setenv("LANG", "C", 1);
    setenv("LC_ALL", "C", 1);
    setenv("LANGUAGE", "C", 1);
    setenv("PATH", "/root/bin:/usr/bin:/bin:/usr/sbin:/sbin", 1);

    setenv();

    if(!is(arch)) {
        struct utsname sysinfo;
        uname(&sysinfo);
        cpu(sysinfo.machine);
    }
    else {
        cpu(*arch);
#ifdef  HAVE_PERSONALITY
        keydata *pk = keyf.get("personality");
        cp = NULL;
        if(pk) {
            cp = pk->get(*arch);
            if(cp && eq(cp, "i386")) {
                setenv("CAPE_PERSONALITY", "linux32", 1);
                pmode = PER_LINUX32;
            }
        }
        else if(eq(*arch, "i386")) {
            setenv("CAPE_PERSONALITY", "linux32", 1);
            pmode = PER_LINUX32;
        }
#endif
    }

    setenv("CAPE_ARCH", *arch, 1);
    if(is(verbose)) {
        unsigned cmdind = 1;
        const char *argv0 = *argv;
        if(eq(argv0, "fakeroot")) {
            argv0 = argv[1];
            ++cmdind;
        }

        while(argv[cmdind] && *argv[cmdind] == '-')
            ++cmdind;
        const char *cmdarg = argv[cmdind];
        if(!cmdarg)
            cmdarg = "";
        printf("executing %s for %s %s\n", prefix, *argv, cmdarg);
    }

#ifdef  HAVE_PERSONALITY
    if(pmode)
        personality(pmode);
#endif
    execvp(*argv, argv);
    ::exit(-1);
}

// cleanup runtime directory of unused mount pounts
static void runtime_cleanup(const char *path)
{
    char buffer[128];
    dir_t dir(path);
    while(is(dir) && dir.read(buffer, sizeof(buffer)) > 0) {
        if(*buffer < '0' || *buffer > '9')
            continue;
        rmdir(buffer);
    }
    dir.close();
}

static void runclean(session_t s)
{
    char *args[16];
    const char *distro = s.distro;

    setenv("DEBIAN_FRONTEND", "noninteractive", 1);
    if(chdir(s.prefix))
        shell::errexit(2, "*** cape-chroot: %s: %s",
            *s.prefix, _TEXT("cannot access directory"));

    string_t yum = str(s.prefix) + "/etc/yum.repos.d";
    if(fsys::is_dir(*yum)) {
        args[0] = (char *)"yum";
        args[1] = (char *)"-y";
        args[2] = (char *)"clean";
        args[3] = (char *)"all";
        args[4] = NULL;
        if(runroot(args))
            shell::errexit(20, "*** cape-chroot: %s: %s",
                distro, _TEXT("yum clean all failed"));
    }
    else {
        args[0] = (char *)"apt-get";
        args[1] = (char *)"-y";
        args[2] = (char *)"clean";
        args[3] = NULL;
        if(runroot(args))
            shell::errexit(20, "*** cape-chroot: %s: %s",
                distro, _TEXT("apt-get clean failed"));
    }
}

static int runupdates(char **argv = NULL, bool cleanup = false)
{
    char *args[256];
    unsigned index = 0;
    const char *distro = session.distro;
    char *prefix = strdup(str(session.prefix) + "/tmp/buildd");

    setenv("DEBIAN_FRONTEND", "noninteractive", 1);
    if(chdir(prefix))
        shell::errexit(2, "*** cape-chroot: %s: %s",
            prefix, _TEXT("cannot access directory"));

    string_t yum = str(session.prefix) + "/etc/yum.repos.d";
    if(fsys::is_dir(*yum)) {
        args[0] = (char *)"yum";
        args[1] = (char *)"-y";
        args[2] = (char *)"update";
        args[3] = NULL;
        index = 2;
        if((!argv || *argv == NULL) && runroot(args))
            shell::errexit(20, "*** cape-chroot: %s: %s",
                distro, _TEXT("yum update failed"));
    }
    else {
        args[0] = (char *)"apt-get";
        args[1] = (char *)"-y";
        args[2] = (char *)"--no-install-recommends";
        args[3] = (char *)"update";
        index = 4;
        if(is(localflag)) {
            args[index++] = (char *)"-o";
            args[index++] = (char *)"Dir::Etc::sourcelist=sources.list.d/chroot.list";
            args[index++] = (char *)"-o";
            args[index++] = (char *)"Dir::Etc::sourceparts=-";
            args[index++] = (char *)"-o";
            args[index++] = (char *)"APT::Get::List-Cleanup=0";
        }
        args[index] = NULL;
        if(runroot(args))
            shell::errexit(20, "*** cape-chroot: %s: %s",
                distro, _TEXT("apt-get update failed"));

        args[2] = (char *)"--no-install-recommends";
        args[3] = (char *)"dist-upgrade";
        args[4] = NULL;
        index = 3;
        if((!argv || *argv == NULL) && runroot(args))
            shell::errexit(20, "*** cape-chroot: %s: %s",
                distro, _TEXT("apt-get dist-upgrade failed"));
    }
    if(!argv || !argv[0])
        goto clean;

    args[index++] = (char *)"install";
    while(*argv)
        args[index++] = *(argv++);
    args[index] = NULL;

    if(runroot(args))
        shell::errexit(20, "*** cape-chroot: %s: %s",
            distro, _TEXT("package installation failed"));

clean:
    if(!cleanup)
        return 0;

    if(fsys::is_dir(*yum)) {
        index = 2;
        if(!is(verbose))
            args[index++] = (char *)"--quiet";
        args[index++] = (char *)"clean";
        args[index++] = (char *)"packages";
        args[index--] = NULL;
        if(runroot(args))
            shell::errexit(20, "*** cape-chroot: %s: %s",
                distro, _TEXT("yum clean packages failed"));
        args[index] = (char *)"headers";
        if(runroot(args))
            shell::errexit(20, "*** cape-chroot: %s: %s",
                distro, _TEXT("yum clean headers failed"));
    }
    else {
        index = 2;
        if(!is(verbose))
            args[index++] = (char *)"-q";
        args[index++] = (char *)"autoremove";
        args[index--] = NULL;
        if(runroot(args))
            shell::errexit(20, "*** cape-chroot: %s: %s",
                distro, _TEXT("apt-get autoremove failed"));
        args[index] = (char *)"autoclean";
        if(runroot(args))
            shell::errexit(20, "*** cape-chroot: %s: %s",
                distro, _TEXT("apt-get autoclean failed"));
    }
    return 0;
}

static int runbuild(char **argv)
{
    bool yumflag = false;
    string_t buildd = str(session.prefix) + "/tmp/buildd";
    char *args[32];
    char jarg[4];
    const char *distro = session.distro;
    const char *ext;

    setenv("DEBIAN_FRONTEND", "noninteractive", 1);
    setenv("LANG", "C", 1);
    setenv("LC_ALL", "C", 1);
    setenv("LANGUAGE", "C", 1);

    string_t yum = str(session.prefix) + "/etc/yum.repos.d";
    if(fsys::is_dir(*yum))
        yumflag = true;

//  runupdates(keyf, s);
    while(argv && *argv) {
        if(chdir(*buildd))
            shell::errexit(2, "*** cape-chroot: %s: %s",
                *buildd, _TEXT("cannot access directory"));
        ext = strrchr(*argv, '.');
        char *pkg = strrchr(*argv, '/');
        if(!pkg)
            pkg = *argv;
        else
            ++pkg;
        if(!ext)
            ext = ".none";
        if(yumflag || eq(ext, ".rpm")) {
            args[0] = (char *)"yum-downloader";
            args[1] = (char *)"--source";
            args[2] = *argv;
            args[3] = NULL;
            if(runroot(args))
                shell::errexit(20, "*** cape-chroot: %s: %s: %s",
                    distro, pkg, _TEXT("yum-downloader failed"));
            if(!is(nodeps)) {
                args[0] = (char *)"yum-builddep";
                args[1] = (char *)"-y";
                args[2] = *argv;
                args[3] = NULL;
                if(runroot(args))
                    shell::errexit(20, "*** cape-chroot: %s: %s: %s",
                        distro, pkg, _TEXT("yum-builddep failed"));
            }
        }
        else {
            if(eq(ext, ".dsc")) {
                dsc_copy(*argv, buildd);
                char *ep;
                string_t deps = "";
                char buffer[256];
                file_t dsc(pkg, "r");
                while(dsc.getline(buffer, sizeof(buffer)) > 0) {
                    if(eq_case(buffer, "Build-Depends", 13))
                        continue;
                    ep = strchr(buffer, ':');
                    if(!ep)
                        continue;
                    ++ep;
                    if(buffer[13] != ':' && (is(binary) || is(arch)))
                        continue;
                    if(deps.len())
                        deps = str(deps) + "," + (const char *)ep;
                    else
                        deps = str(ep);
                }
                dsc.close();

                args[0] = (char *)"dpkg-source";
                args[1] = (char *)"-x";
                args[2] = (char *)strdup(pkg);
                args[3] = NULL;
                ep = strchr(pkg, '_');
                *ep = 0;

                if(runroot(args))
                    shell::errexit(20, "*** cape-chroot: %s: %s: %s",
                        distro, pkg, _TEXT("unpack source failed"));

                if(deps.len() > 0 && !is(nodeps)) {
                    fsys::erase("chroot-build-deps_1.0_all.deb");
                    FILE *eq = fopen("equiv", "w");
                    fprintf(eq,
                        "Section: devel\n"
                        "Priority: optional\n"
                        "Standards-Version: 3.7.3\n"
                        "Package: chroot-build-deps\n"
                        "Depends: %s\n"
                        "Description: build-dependencies for %s\n"
                        " Dependency package to build in chroot\n", *deps, pkg);
                    fclose(eq);
                    args[0] = (char *)"equivs-build";
                    args[1] = (char *)"equiv";
                    args[2] = NULL;
                    if(runuser(args, "/tmp/buildd"))
                        shell::errexit(20, "*** cape-chroot: %s: %s: %s",
                            distro, pkg, _TEXT("cannot create build-deps"));
                    args[0] = (char *)"dpkg";
                    args[1] = (char *)"--unpack";
                    args[2] = (char *)"chroot-build-deps_1.0_all.deb";
                    args[3] = 0;
                    if(runroot(args))
                        shell::errexit(20, "*** cape-chroot: %s: %s: %s",
                            distro, pkg, _TEXT("unpacking dependencies failed"));
                    args[0] = (char *)"apt-get";
                    args[1] = (char *)"-fy";
                    args[2] = (char *)"install";
                    args[3] = NULL;
                    if(runroot(args))
                        shell::errexit(20, "*** cape-chroot: %s: %s: %s",
                            distro, pkg, _TEXT("installing dependencies failed"));

                    args[0] = (char *)"dpkg";
                    args[1] = (char *)"-r";
                    args[2] = (char *)"chroot-build-deps";
                    args[3] = 0;
                    if(runroot(args))
                        shell::errexit(20, "*** cape-chroot: %s: %s: %s",
                            distro, pkg, _TEXT("removing dependencies failed"));
                }
                fsys::erase("chroot-build-deps_1.0_all.deb");
            }
            else {
                args[0] = (char *)"apt-get";
                args[1] = (char *)"-y";
                args[2] = (char *)"source";
                args[3] = *argv;
                args[4] = NULL;
                if(runroot(args))
                    shell::errexit(20, "*** cape-chroot: %s: %s: %s",
                        distro, pkg, _TEXT("apt-get source failed"));
                args[2] = (char *)"build-dep";
                if(runroot(args))
                    shell::errexit(20, "*** cape-chroot: %s: %s: %s",
                        distro, pkg, _TEXT("apt-get build-dep failed"));
            }
            if(!sourcedir(str(pkg) + "-"))
                shell::errexit(20, "*** cape-chroot: %s: %s: %s",
                    distro, pkg, _TEXT("failed to unpack source"));
            args[0] = (char *)"dpkg-buildpackage";
            if(is(binary) || is(arch))
                args[1] = (char *)"-B";
            else
                args[1] = (char *)"-b";
            args[2] = (char *)"-uc";
            if(*jobs) {
                snprintf(jarg, sizeof(jarg), "-j%ld", *jobs);
                args[3] = jarg;
                args[4] = NULL;
            }
            else
                args[3] = NULL;
            if(runroot(args))
                shell::errexit(20, "*** cape-chroot: %s: %s: %s",
                    distro, pkg, _TEXT("build failed"));
            if(chdir(buildd))
                shell::errexit(2, "*** cape-chroot: %s: %s",
                    *buildd, _TEXT("cannot access directory"));
            deb_results(str(pkg) + "_");
        }
        ++argv;
    }
    return 0;
}

static void cleanup(void)
{
    static bool clean = false;

    if(getpid() == xpid)
        return;

    if(xpid) {
        killpg(xpid, SIGKILL);
        waitpid(xpid, NULL, 0);
        xpid = 0;
    }

    if(clean || getpid() != exitpid)
        return;

    clean = true;

    fflush(stdout);
    fflush(stderr);

    // detach for release...
    if(!is(verbose))
        shell::release(exit_code);

    if(chdir("/"))
        shell::errexit(20, "*** cape-chroot: %s",
            _TEXT("cannot access root prefix"));

    sync();

    // unmount in reverse order...

    while(mcount-- > 0) {
        if(is(verbose))
            printf("unmounting %s\n", mtable[mcount]);
        umount(mtable[mcount]);
        sync();
    }

    if(is(verbose))
        printf("releasing session %s\n", session.id);

    rmdir(*session.prefix);
    ::exit(exit_code);
}

static void stop(int signo)
{
    exit_code = signo;
    cleanup();
}

static int startup(void *arg)
{
    const char *cp = NULL, *archive, *backports, *common;
    keydata *keys, *sys = keyf.get("system");
    char *token, *tokens;
    FILE *fp;
    const char *distro = session.distro;
    const char *prefix = session.prefix;
    const char *origin = session.origin;
    string_t tmp;
    int mask;
    int result = 0;

#ifdef  CLONE_NEWPID
    bool cloned = true;
#else
    bool cloned = false;
#endif

    // mounts auto-disconnect when process ends...
#ifdef  HAVE_UNSHARE
    unshare(CLONE_NEWNS);
#endif

    if(!is(results)) {
        if(sys)
            cp = sys->get("results");
        if(cp)
            results.set(cp);
        else
            results.set(strdup(shell::path(shell::SYSTEM_VAR) + "/lib/chroot/results"));
        mkdir(*results, 0775);
    }

    if(sys)
        cp = sys->get("archive");
    if(cp)
        archive = strdup(cp);
    else
        archive = strdup(shell::path(shell::SYSTEM_VAR) + "/lib/chroot/archive");

    if(sys)
        cp = sys->get("common");
    if(cp)
        common = strdup(cp);
    else
        common = strdup(shell::path(shell::SYSTEM_VAR) + "/lib/chroot/common");

    if(sys)
        cp = sys->get("backports");
    if(cp)
        backports = strdup(cp);
    else
        backports = strdup(shell::path(shell::SYSTEM_VAR) + "/lib/chroot/backports");



    if(fsys::is_dir(origin)) {
        keys = keyf.get("copy");
        if(!keys || NULL != (cp = keys->get(distro))) {
            if(sys)
                cp = sys->get("copy");
            else
                cp = NULL;

            if(!cp)
                cp = "/etc/shadow /etc/gshadow /etc/group /etc/passwd /etc/sudoers /etc/hosts /etc/resolv.conf /etc/hostname";
        }

        if(cp) {
            tokens = strdup(cp);
            token = strtok(tokens, " \t\n,;:");
        }
        else
            token = NULL;

        // never copy things into a system mode chroot...

        while(!is(live) && token) {
            if(*token == '/' && fsys::is_file(token)) {
                string_t from = token;
                string_t to = str(origin) + (const char *)token;
                copyfile(from, to);
            }
            token = strtok(NULL, " \t\n,;:");
        }

        if(mount(origin, prefix, NULL, MS_BIND, NULL))
            shell::errexit(6, "*** cape-chroot: %s: %s\n",
                session.id, _TEXT("session failed"));
        if(is(verbose))
            printf("mounting %s\n", *session.prefix);
        mtable[mcount++] = strdup(session.prefix);
    }
    else {
        rmdir(prefix);
        shell::errexit(7, "*** cape-chroot: %s: %s\n",
            origin, _TEXT("chroot not found"));
    }

    tmp = str(prefix) + "/root";
    string_t common_root = str(common) + "/root";
    if(fsys::is_dir(*common_root)) {
        if(is(verbose))
            printf("mounting %s\n", *tmp);
        mkdir(*tmp, 0775);
        mount(*common_root, *tmp, NULL, MS_BIND, NULL);
        mtable[mcount++] = strdup(*tmp);
    }

    // for building mount tmp into ramdisk...
    if(root || is(update) || is(clean) || is(build) || is(binary)) {
        mask = umask(0);
        tmp = str(prefix) + "/tmp";
        if(is(verbose))
            printf("mounting %s\n", *tmp);
        mkdir(*tmp, 01777);
        mount("none", *tmp, "tmpfs", 0, "mode=01777");
        mtable[mcount++] = strdup(*tmp);
        tmp = str(prefix) + "/tmp/buildd";
        mkdir(*tmp, 01777);
        umask(mask);
    }
    // mount tmp needed for X socket...
    else {
        mask = umask(0);
        tmp = str(prefix) + "/tmp";
        if(is(verbose))
            printf("mounting %s\n", *tmp);
        mkdir(*tmp, 01777);
        mount("/tmp", *tmp, NULL, MS_BIND, NULL);
        mtable[mcount++] = strdup(*tmp);
        umask(mask);
    }

    // var tmp expedient...
    tmp = str(prefix) + "/var/tmp";
    if(is(verbose))
        printf("mounting %s\n", *tmp);
    mask = umask(0);
    mkdir(*tmp, 01777);
    mount("none", *tmp, "tmpfs", 0, "mode=01777");
    mtable[mcount++] = strdup(*tmp);
    umask(mask);

    if(!is(live)) {
        // create working directories...
        mkdir(str(archive) + "/" + distro, 0775);
        mkdir(str(archive) + "/source", 0775);

        // mount var archive
        tmp = str(prefix) + "/var/archive";
        if(is(verbose))
            printf("mounting %s\n", *tmp);
        mkdir(*tmp, 0775);
        mount(archive, *tmp, NULL, MS_BIND, NULL);
        mtable[mcount++] = strdup(*tmp);

        // setup repository support
        tmp = str(prefix) + "/etc/apt/sources.list.d";
        if(fsys::is_dir(*tmp)) {
            string_t listfile = str(tmp) + "/chroot.list";
            string_t sources = str(archive) + "/source/Sources";
            string_t backport = str(archive) + "/" + distro + "/Sources";
            string_t packages = str(archive) + "/" + distro + "/Packages";
            fp = fopen(*listfile, "w");
            if(fp) {
                if(is(verbose))
                    printf("creating %s\n", *listfile);
                if(fsys::is_file(*packages))
                    fprintf(fp, "deb file:///var/archive/ %s/\n", distro);
                if(fsys::is_file(*backport))
                    fprintf(fp, "deb-src file:///var/archive/ %s/\n", distro);
                else if(fsys::is_file(*sources))
                    fprintf(fp, "deb-src file:///var/archive/ source/\n");
                fclose(fp);
            }
            else
                fprintf(stderr, "*** cape-chroot: %s: %s\n",
                    *tmp, _TEXT("failed to create"));
        }

        // create backports...
        if(fsys::is_dir(backports)) {
            mkdir(str(backports) + "/" + distro, 0775);

            tmp = str(prefix) + "/var/backports";
            if(is(verbose))
                printf("mounting %s\n", *tmp);
            mkdir(*tmp, 0775);
            mount(backports, *tmp, NULL, MS_BIND, NULL);
            mtable[mcount++] = strdup(*tmp);

            // setup backports support
            tmp = str(prefix) + "/etc/apt/sources.list.d";
            if(fsys::is_dir(*tmp)) {
                string_t listfile = str(tmp) + "/backports.list";
                string_t packages = str(backports) + "/" + distro + "/Packages";
                fp = fopen(*listfile, "w");
                if(fp) {
                    if(is(verbose))
                        printf("creating %s\n", *listfile);
                    if(fsys::is_file(*packages))
                        fprintf(fp, "deb file:///var/backports/ %s/\n", distro);
                    fclose(fp);
                }
                else
                    fprintf(stderr, "*** cape-chroot: %s: %s\n",
                        *tmp, _TEXT("failed to create"));
            }
        }

        tmp = str(prefix) + "/etc/yum.repos.d";
        if(fsys::is_dir(*tmp)) {
            string_t listfile = str(tmp) + "/chroot.repo";
            string_t sources = str(archive) + "/source/repodata";
            string_t packages = str(archive) + "/" + distro + "/repodata";
            fp = fopen(*listfile, "w");
            if(fp) {
                if(is(verbose))
                    printf("creating %s\n", *listfile);
                if(fsys::is_dir(*packages))
                    fprintf(fp, "[chroot-binary]\n"
                                "name = chroot binary packages\n"
                                "baseurl = file:///var/archive/%s/\n\n", distro);
                if(fsys::is_file(*sources))
                    fprintf(fp, "[chroot-source]\n"
                                "name = chroot binary packages\n"
                                "baseurl = file:///var/archive/source/\n\n");
                fclose(fp);
            }
            else
                fprintf(stderr, "*** cape-chroot: %s: %s\n",
                    *tmp, _TEXT("failed to create"));
        }
    }

    // proc
    tmp = str(prefix) + "/proc";
    if(fsys::is_dir(*tmp)) {
        if(is(verbose))
            printf("mounting %s\n", *tmp);
#ifdef  CLONE_NEWPID
        mount("proc", *tmp, "proc", MS_MGC_VAL, NULL);
#else
        mount("none", *tmp, "proc", 0, NULL);
#endif
        mtable[mcount++] = strdup(*tmp);
    }

    // sys
    tmp = str(prefix) + "/sys";
    if(fsys::is_dir(*tmp)) {
        if(is(verbose))
            printf("mounting %s\n", *tmp);
        mount("none", *tmp, "sysfs", 0, NULL);
        mtable[mcount++] = strdup(*tmp);
    }

    if(!is(update) && !is(clean) && !root && !is(build) && !is(binary)) {
        const char *varrun = "/var/run";
        if(fsys::is_dir("/run"))
            varrun = "/run";

        // give chroot access to run directory...
        tmp = str(prefix) + "/run";
        if(!fsys::is_dir(*tmp))
            tmp = str(prefix) + "/var/run";
        if(is(verbose))
            printf("mounting %s\n", *tmp);
        mkdir(*tmp, 0755);
        mount(varrun, *tmp, NULL, MS_BIND, NULL);
        mtable[mcount++] = strdup(*tmp);

        keys = keyf.get("mount");
        if(!keys || NULL != (cp = keys->get(distro))) {
            if(sys)
                cp = sys->get("mount");
            else
                cp = NULL;
        }
    }

    if(is(xserver) || is(xscreen)) {
        tokens = strdup("/home /dev /dev/pts /dev/shm /var/lib/dbus /var/run/dbus");
        token = strtok(tokens, " \t\n,;:");
        tmp = str(prefix) + "/var/run/dbus";
        mkdir(*tmp, 0755);
    }
    else if(is(update) || root || is(clean) || is(build) || is(binary)) {
        tmp = str(prefix) + "/dev/pts";
        mkdir(*tmp, 0755);
        if(is(verbose))
            printf("mounting %s\n", *tmp);
        mount("none", *tmp, "devpts", 0, NULL);
        mtable[mcount++] = strdup(*tmp);
        tokens = strdup("");
        token = NULL;
    }
    else if(cp) {
        tokens = strdup(sys->get("mount"));
        token = strtok(tokens, " \t\n,;:");
    }
    else {
        tokens = strdup("/home /dev /dev/pts");
        token = strtok(tokens, " \t\n,;:");
    }

    while(token && *token == '/') {
        if(*token == '/' && fsys::is_dir(token)) {
            string_t from = token;
            string_t to = str(prefix) + (const char *)token;
            if(fsys::is_dir(to)) {
                if(is(verbose))
                    printf("mounting %s\n", *to);
                if(mount(*from, *to, NULL, MS_BIND, NULL))
                    fprintf(stderr, "*** cape-chroot: %s: %s\n",
                        *from, _TEXT("unable to mount"));
                else
                    mtable[mcount++] = strdup(*to);
            }
        }
        token = strtok(NULL, " \t\n,;:");
    }
    
    tmp = str(prefix) + str(pwd->pw_dir);
    if(is(home)) {
        // mount alternate home
        tmp = str(prefix) + (const char *)(pwd->pw_dir);
        string_t alt = str(pwd->pw_dir) + "/" + *home;
        if(is(verbose))
            printf("mounting %s\n", *tmp);
        mkdir(*tmp, 0700);
        mount(*alt, *tmp, NULL, MS_BIND, NULL);
        mtable[mcount++] = strdup(*tmp);
    }
    // auto-remap of parts of home dir for chroot filespaces
    else if(fsys::is_dir(*tmp)) {
        string_t localdir = str(pwd->pw_dir) + "/.chroot/" + distro;
        dir_t remaps;
        char fname[65];
        remaps.open(*localdir);
        while(is(remaps) && remaps.read(fname, 65) > 0) {
            if(*fname == '.')
                continue;
            string_t path = localdir + "/" + str(fname);
            tmp = str(prefix) + str(pwd->pw_dir) + "/" + str(fname);
            if(!fsys::is_dir(*tmp) || !fsys::is_dir(*path))
                continue;
            if(is(verbose))
                printf("mounting %s\n", *tmp);
            mount(*path, *tmp, NULL, MS_BIND, NULL);
            mtable[mcount++] = strdup(*tmp);
        }
    }
    if(is(setup))
        exit(0);

    // based on unshare namespace, where mounts becomes private to
    // local task tree...

    if(is(detach) && !is(update) && !is(build) && !is(binary) && !is(clean))
        rmdir(*session.prefix);
    else {
        signal(SIGINT, stop);
        exitpid = getpid();
        atexit(cleanup);
    }

    if(is(clean))
        runclean(session);

    if(is(build) || is(binary))
        result = runbuild(session.argv);
    else if(is(update))
        result = runupdates(session.argv, !is(clean));
    else if(!is(clean)) {
        if(is(newdir)) 
            result = runuser(session.argv, *newdir, cloned);
        else
            result = runuser(session.argv, cwd, cloned);
    }
    return result;
}

PROGRAM_MAIN(argc, argv)
{
    const char *member, *cp = NULL, *rootdir;
    int result = 0;

    shell::bind("cape-chroot");
    shell args(argc, argv);
    argv = args.argv();

    keyf.load(_STR(shell::path(shell::SYSTEM_CFG) + "/chroot.conf"));
    sys = keyf.get("system");
    if(sys)
        cp = sys->get("chroots");
    if(cp)
        rootdir = cp;
    else
        rootdir = strdup(shell::path(shell::SYSTEM_VAR) + "/lib/chroot/system");

    if(is(verbose))
        printf("cape-chroot root prefix %s\n", rootdir);

    fsys::prefix(cwd, sizeof(cwd));

    if(is(helpflag) || is(althelp)) {
        printf("%s: cape-chroot %s\n", _TEXT("Usage"),
            _TEXT("[options] distro [command...]"));
        printf("%s\n\n", _TEXT("Execute distro in a chroot"));
        printf("%s\n", _TEXT("Options:"));
        shell::help();
        printf("\n%s\n", _TEXT("Report bugs to dyfet@gnu.org"));
        PROGRAM_EXIT(0);
    }

    if(!*argv) {
        shell::errexit(9, "*** cape-chroot: %s\n",
            _TEXT("no cape-chroot distro specified"));
    }

    if(geteuid())
        shell::errexit(1, "*** cape-chroot: %s\n",
            _TEXT("requires sudo or setuid"));

    session.distro = *(argv++);
    uid = getuid();
    gid = getgid();

    if(uid)
        root = false;
    else {
        cp = getenv("SUDO_UID");
        if(cp)
            uid = atoi(cp);
        root = true;
        setenv("LANG", "C", 1);
        setenv("LC_ALL", "C", 1);
        setenv("LANGUAGE", "C", 1);
    }

    if(root && *argv)
        shell::errexit(1, "*** cape-chroot: %s\n",
            _TEXT("root access only for login"));

    if(is(arch))
        cpu(*arch);

    pwd = getpwuid(uid);
    grp = getgrnam("chroot");
    if(!grp) {
        shell::errexit(3, "*** chroot: %s\n",
            _TEXT("group entry missing"));
    }

    if(uid) {
        if(!pwd) {
            shell::errexit(2, "*** cape-chroot: %s\n",
                _TEXT("unknown user"));
        }

        gid = pwd->pw_gid;
        unsigned gind = 0;
        while((member = grp->gr_mem[gind++]) != NULL) {
            if(eq(member, pwd->pw_name))
                break;
        }

        if(!gid) {
            shell::errexit(2, "*** cape-chroot: %s: %s\n",
                pwd->pw_name, _TEXT("not member of chroot group"));
        }
    }

    if(is(home)) {
        fsys::fileinfo_t info;
        string_t alt = str(pwd->pw_dir) + "/" + *home;
        if(fsys::info(*alt, &info) || !S_ISDIR(info.st_mode) || info.st_uid != pwd->pw_uid)
            shell::errexit(127, "*** cape-chroot: %s: %s\n",
                *home, _TEXT("access violation"));
        if(chdir(pwd->pw_dir))
            shell::errexit(127, "*** cape-chroot: %s: %s\n",
                pwd->pw_dir, _TEXT("cannot access"));
        String::set(cwd, sizeof(cwd), pwd->pw_dir);
    }

    setenv("HOME", pwd->pw_dir, 1);
    setenv("USER", pwd->pw_name, 1);
    setenv("USERNAME", pwd->pw_name, 1);
    setenv("LOGNAME", pwd->pw_name, 1);

    session.origin = str(rootdir) + "/" + session.distro;
    if(is(arch))
        session.origin = str(session.origin) + "-" + *arch;

    unsigned opts = 0;

    if(is(setup))
        ++opts;

    if(opts > 1)
        shell::errexit(9, "*** cape-chroot: %s",
            _TEXT("conflicting options"));

    mkdir("/tmp/.chroot", 0755);
    if(!fsys::is_dir("/tmp/.chroot") || chdir("/tmp/.chroot")) {
        shell::errexit(8, "*** cape-chroot: %s",
            _TEXT("cannot create runtime directory"));
    }
    runtime_cleanup("/tmp/.chroot");

    snprintf(session.id, sizeof(session.id), "%lu", (unsigned long)getpid());
    session.prefix = "/tmp/.chroot/";
    session.prefix += session.id;
    session.argv = argv;

    if(mkdir(session.prefix, 0755)) {
        shell::errexit(5, "*** cape-chroot: %s\n",
            _TEXT("cannot create session"));
    }

    if(is(verbose))
        printf("creating session %s for %s\n", session.id, *session.distro);

#ifdef  CLONE_NEWPID
    const pid_t cpid = clone(startup, cstack + STACK_SIZE, CLONE_NEWNS | CLONE_NEWPID | SIGCHLD, NULL);
    if(cpid == -1)
        shell::errexit(9, "*** cape-chroot: %s\n",
            _TEXT("cannot clone"));
    waitpid(cpid, &result, 0);
    if(WIFEXITED(result))
        result = WEXITSTATUS(result);
    else
        result = -1;
#else
    result = startup(NULL);
#endif
    return result;
}
