// Copyright (C) 2012 David Sugar, Tycho Softworks.
//
// This file is part of  cape-devtools.
//
// Cape-devtools is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Cape-devtools is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Cape-devtools.    If not, see <http://www.gnu.org/licenses/>.

#include <cape-devtools-config.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/wait.h>
#include <sys/utsname.h>
#include <sys/fcntl.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <ucommon/ucommon.h>

#ifdef  HAVE_UNSHARE
#include <sched.h>
#endif

#pragma GCC diagnostic ignored "-Wunused-result"

using namespace UCOMMON_NAMESPACE;

static shell::flagopt helpflag('h',"--help",    _TEXT("display this list"));
static shell::flagopt althelp('?', NULL, NULL);
static shell::flagopt detach('d', "--detach", _TEXT("detach child process"));
static shell::flagopt force('f', "--force", _TEXT("force copy of key files"));
static shell::flagopt verbose('v', "--verbose", _TEXT("verbose logging"));

static bool clean_flag = false;
static int exit_code = 0;
static char *mtable[64];
static unsigned mcount;
static char cwd[256];
static char sid[40];
static struct passwd *pwd = NULL;
static struct group *grp = NULL;
static int uid, gid;
static const char *home;
static char **sys_argv;

#ifdef  CLONE_NEWPID
#define STACK_SIZE (1024 * 1024)
static char cstack[STACK_SIZE];
#endif

static void copyfile(string_t from, string_t to)
{
    char buffer[4096];

    fd_t source, target;
    fsys::fileinfo_t finfo, tinfo;
    string_t tmp = str(to) + ".tmp.";
    if(fsys::info(*from, &finfo))
        return;

    // if file unchanged, no need to copy
    if(!fsys::info(*to, &tinfo)) {
        if(!is(force) && finfo.st_mtime < tinfo.st_mtime && finfo.st_size == tinfo.st_size)
            return;
    }

    source = open(*from, O_RDONLY);
    if(source < 0)
        return;

    if(is(force))
        remove(*tmp);

    // if another session already doing it, ignore...
    target = open(*tmp, O_CREAT | O_EXCL | O_WRONLY, finfo.st_mode & 0777);
    if(target < 0) {
        close(source);
        return;
    }

    if(is(verbose))
        printf("copying %s\n", *from);

    for(;;) {
        ssize_t len = read(source, buffer, sizeof(buffer));
        if(len < 0) {
            close(source);
            close(target);
            remove(*tmp);
            fprintf(stderr, "*** cape-chroot: %s: %s\n",
                *from, _TEXT("unable to copy"));
            exit(0);
        }
        if(len > 0)
            if(write(target, buffer, len) < len)
                shell::errexit(10, "*** cape-chroot: %s",
                    _TEXT("cannot copy"));
        if((size_t)len < sizeof(buffer))
            break;
    }
    close(source);
    close(target);
    rename(*tmp, *to);
    if(chown(*to, 0, 0))
        printf("*** warning: %s: cannot change owner", *to);
}

static int runuser(char **argv)
{
    const char *cmd = *argv;

#ifndef  CLONE_NEWPID
    int status;
    if(!is(detach)) {
        pid_t pid = fork();
        if(pid) {
            waitpid(pid, &status, 0);
            if(WIFEXITED(status))
                exit_code = WEXITSTATUS(status);
            else
                exit_code = -1;
            return exit_code;
        }
    }
#endif

    setgid(gid);
    setegid(gid);
    setuid(uid);
    seteuid(uid);

    const char *shell = getenv("SHELL");

    if(!shell) {
        shell = "/bin/bash";
        setenv("SHELL", "/bin/bash", 1);
    }

    if(!chdir(pwd->pw_dir)) {
        if(is(verbose))
            printf("changing to %s\n", pwd->pw_dir);
    }

    setenv("CAPE_SESSION", sid, 1);
    setenv("CAPE_DISTRIBUTION", "native", 1);
    setenv("CAPE_PERSONALITY", "native", 1);
    setenv("CAPE_ARCH", "native", 1);

    if(!*argv) {
        char *args[3];
        args[0] = (char *)shell;
        args[1] = (char *)"-l";
        args[2] = NULL;
        argv = args;
        if(is(verbose))
            printf("logging into %s\n", pwd->pw_dir);
        cmd = shell;
    }
    else {
        if(is(verbose))
            printf("executing %s\n", pwd->pw_dir);
        cmd = *argv;
    }
    execvp(cmd, argv);
    fprintf(stderr, "*** cape-rehome: %s: %s\n",
        cmd, _TEXT("failed to run"));
    if(!is(detach)) // block duplicate clean from forked instance...
        clean_flag = true;
    exit(-1);
}

// cleanup runtime directory of unused mount pounts

static void cleanup(void)
{
    if(clean_flag)
        return;

    clean_flag = true;

    fflush(stdout);
    fflush(stderr);

    // detach for release...
    if(!is(verbose))
        shell::release(exit_code);

    if(chdir("/"))
        shell::errexit(20, "*** cape-rehome: %s",
            _TEXT("cannot access root prefix"));

    sync();

    // unmount in reverse order...

    while(mcount-- > 0) {
        if(is(verbose))
            printf("unmounting %s\n", mtable[mcount]);
        umount(mtable[mcount]);
        sync();
    }

    if(is(verbose))
        printf("releasing session %s\n", sid);

    ::exit(exit_code);
}

static void stop(int signo)
{
    exit_code = signo;
    cleanup();
}

static int startup(void *arg)
{
    string_t tmp;

    // mounts auto-disconnect when process ends...
#ifdef  HAVE_UNSHARE
    unshare(CLONE_NEWNS);
#endif

    static const char *list[] = {"/.Xauthority", "/.profile", "/.bashrc", NULL};
    unsigned index = 0;
    while(list[index]) {
        string_t from = str(pwd->pw_dir) + list[index];
        string_t to = str(home) + list[index];
        copyfile(from, to);
        ++index;
    }

    if(is(verbose))
        printf("mounting %s as %s\n", home, pwd->pw_dir);
    mount(home, pwd->pw_dir, NULL, MS_BIND, NULL);
    mtable[mcount++] = strdup(pwd->pw_dir);

    // based on unshare namespace, where mounts becomes private to
    // local task tree...

    if(!is(detach)) {
        signal(SIGINT, stop);
        atexit(cleanup);
    }

    runuser(sys_argv);
    return 0;
}

PROGRAM_MAIN(argc, argv)
{
    const char *member, *cp = NULL;

    shell::bind("cape-rehome");
    shell args(argc, argv);
    sys_argv = args.argv();

    if(is(helpflag) || is(althelp)) {
        printf("%s: cape-rehome %s\n", _TEXT("Usage"),
            _TEXT("[options] homedir [command...]"));
        printf("%s\n\n", _TEXT("Execute distro in a chroot"));
        printf("%s\n", _TEXT("Options:"));
        shell::help();
        printf("\n%s\n", _TEXT("Report bugs to dyfet@gnu.org"));
        PROGRAM_EXIT(0);
    }

    if(!*sys_argv) {
        shell::errexit(9, "*** cape-rehome: %s\n",
            _TEXT("no cape-rehome directory specified"));
    }

    if(geteuid())
        shell::errexit(1, "*** cape-rehome: %s\n",
            _TEXT("requires sudo or setuid"));

    home = *(sys_argv++);
    uid = getuid();
    gid = getgid();

    if(!uid) {
        cp = getenv("SUDO_UID");
        if(cp)
            uid = atoi(cp);
    }

    pwd = getpwuid(uid);
    grp = getgrnam("chroot");
    if(!grp) {
        shell::errexit(3, "*** chroot: %s\n",
            _TEXT("group entry missing"));
    }

    if(uid) {
        if(!pwd) {
            shell::errexit(2, "*** cape-rehome: %s\n",
                _TEXT("unknown user"));
        }

        gid = pwd->pw_gid;
        unsigned gind = 0;
        while((member = grp->gr_mem[gind++]) != NULL) {
            if(eq(member, pwd->pw_name))
                break;
        }

        if(!gid) {
            shell::errexit(2, "*** cape-rehome: %s: %s\n",
                pwd->pw_name, _TEXT("not member of chroot group"));
        }
    }

    fsys::fileinfo_t info;
    string_t alt;

    if(*home == '/')
        alt = home;
    else
        alt = str(pwd->pw_dir) + "/" + home;

    if(fsys::info(*alt, &info) || !S_ISDIR(info.st_mode) || info.st_uid != pwd->pw_uid)
        shell::errexit(127, "*** cape-rehome: %s: %s\n",
                home, _TEXT("access violation"));

    if(chdir(pwd->pw_dir))
        shell::errexit(127, "*** cape-rehome: %s: %s\n",
            pwd->pw_dir, _TEXT("cannot access"));
    String::set(cwd, sizeof(cwd), pwd->pw_dir);

    setenv("HOME", pwd->pw_dir, 1);
    setenv("USER", pwd->pw_name, 1);
    setenv("USERNAME", pwd->pw_name, 1);
    setenv("LOGNAME", pwd->pw_name, 1);

#ifdef  CLONE_NEWPID
    const pid_t cpid = clone(startup, cstack + STACK_SIZE, CLONE_NEWNS | CLONE_NEWPID | SIGCHLD, NULL);
    if(cpid == -1)
        shell::errexit(9, "*** cape-chroot: %s\n",
            _TEXT("cannot clone"));
    waitpid(cpid, NULL, 0);
#else
    startup(NULL);
#endif
}
