// Copyright (C) 2012 David Sugar, Tycho Softworks.
//
// This file is part of  cape-devtools.
//
// Cape-devtools is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Cape-devtools is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Cape-devtools.    If not, see <http://www.gnu.org/licenses/>.

#include <cape-devtools-config.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/wait.h>
#include <sys/utsname.h>
#include <sys/fcntl.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <ucommon/ucommon.h>

using namespace UCOMMON_NAMESPACE;

static shell::flagopt helpflag('h',"--help",    _TEXT("display this list"));
static shell::flagopt althelp('?', NULL, NULL);
static shell::flagopt changes('c', "--changes", _TEXT("generate changes file"));
static shell::flagopt lint('l', "--lint", _TEXT("lint check packages"));
static shell::stringopt profile('p', "--profile", _TEXT("project profile"), "config", "default");
static shell::flagopt sign('s', "--sign", _TEXT("sign index"));
static shell::flagopt verbose('v', "--verbose", _TEXT("verbose"));
static shell::flagopt who('w', "--who", _TEXT("who am i"));

static keydata *archives = NULL;
static char prefix[512];
static const char *tarbase, *version, *format, *tarball, *tarfile, *revision;
static keyfile pkgbuildrc;
static pid_t pid;
static const char *arch = NULL, *releaseid = NULL;
static const char *cleandir = NULL, *label = NULL, *origin = NULL, *distribution = "unstable";
static const char *realname, *email, *signid = NULL, *vendor = NULL;
static bool debianos = false, redhatos = false;

static void copyfile(String& from, String& to)
{
    char buffer[4096];

    fd_t source, target;
    fsys::fileinfo_t finfo, tinfo;
    String tmp = str(to) + ".tmp.";
    if(fsys::info(*from, &finfo))
        return;

    // if file unchanged, no need to copy
    if(!fsys::info(*to, &tinfo)) {
        if(finfo.st_mtime < tinfo.st_mtime && finfo.st_size == tinfo.st_size)
            return;
    }

    source = open(*from, O_RDONLY);
    if(source < 0)
        return;

    remove(*tmp);

    // if another session already doing it, ignore...
    target = open(*tmp, O_CREAT | O_EXCL | O_WRONLY, finfo.st_mode & 0777);
    if(target < 0) {
        close(source);
        return;
    }

    if(is(verbose))
        printf("copying %s\n", *from);

    for(;;) {
        ssize_t len = read(source, buffer, sizeof(buffer));
        if(len < 0) {
            close(source);
            close(target);
            remove(*tmp);
            fprintf(stderr, "*** cape-source: %s: %s\n",
                *from, _TEXT("unable to copy"));
        }
        if(len > 0) {
            if(write(target, buffer, len) < len)
                shell::errexit(10, "*** cape-source: %s",
                    _TEXT("failed to copy"));
        }
        if((size_t)len < sizeof(buffer))
            break;
    }
    close(source);
    close(target);
    rename(*tmp, *to);
}

static void erasefiles(const char *path, const char *prefix, const char *suffix)
{
    char buffer[65];
    dir_t dir(path);
    size_t plen = strlen(prefix), slen = strlen(suffix), len;
    while(is(dir) && dir.read(buffer, sizeof(buffer)) > 0) {
        len = strlen(buffer);
        if(len <= slen)
            continue;
        if(!eq(prefix, buffer, plen))
            continue;
        if(!eq(suffix, buffer + len - slen))
            continue;
        if(is(verbose))
            printf("removing %s\n", buffer);
        fsys::erase(str(path) + "/" + (const char *)buffer);
    }
}

static void check_changes(const char *changelog)
{
    char change[96];
    file_t chfile(changelog, "r");
    chfile.getline(change, sizeof(change));

    char *ver = strchr(change, '(');
    char *ep = NULL;

    if(ver) {
        ++ver;
        ep = strchr(ver, ')');
        if(ep)
            *ep = 0;
    }
    if(ver)
        ep = strchr(ver, '-');
    if(ep) {
        *ep = 0;
        revision = strdup(++ep);
    }
    else
        revision = NULL;

    if(!eq(ver, version) || !revision)
        shell::errexit(3, "*** cape-source: %s: %s\n",
            changelog, _TEXT("invalid changes version"));
}

static void check_control(const char *control)
{
    file_t spec(control, "r");
    char buffer[128];
    const char *cp;

    while(spec.getline(buffer, sizeof(buffer))) {
        if(eq_case(buffer, "Source:", 7)) {
            cp = buffer + 7;
            while(isspace(*cp))
                ++cp;
            if(!eq(cp, tarbase))
                shell::errexit(3, "*** cape-source: %s: %s\n",
                    control, _TEXT("invalid source name"));
            else
                return;
        }
    }
    shell::errexit(3, "*** cape-source: %s: %s\n",
        control, _TEXT("invalid control file"));
}

static void check_spec(const char *specfile)
{
    bool okv = false, oks = false;
    file_t spec(specfile, "r");
    char buffer[128];
    const char *cp;

    while(spec.getline(buffer, sizeof(buffer))) {
        if(eq_case(buffer, "Version:", 8)) {
            cp = buffer + 8;
            while(isspace(*cp))
                ++cp;
            if(!eq(cp, version))
                shell::errexit(3, "*** cape-source: %s: %s\n",
                    specfile, _TEXT("version mismatch with source"));
            okv = true;
        }
        if(eq_case(buffer, "Name:", 5)) {
            cp = buffer + 5;
            while(isspace(*cp))
                ++cp;
            if(eq(cp, tarbase))
                oks = true;
        }
    }
    if(!okv || !oks)
        shell::errexit(3, "*** cape-source: %s: %s\n",
            specfile, _TEXT("spec file invalid"));
}

static int deb_lintian(const char *dsc)
{
    char *argv[16];
    unsigned index = 0;

    if(!is(lint))
        return 0;

    argv[index++] = (char *)"lintian";
    if(is(verbose))
        argv[index++] = (char *)"-I";
    argv[index++] = (char *)dsc;
    argv[index++] = NULL;

    return fsys::exec(*argv, argv);
}

// easiest case...self contained tarball with spec...
static void rpm_tarball(void)
{
    char *argv[24];
    unsigned index = 0;
    String srcdir = str("_srcrpmdir ") + (const char *)prefix;
    String packager = str("_packager ") + (const char *)realname + " <" + (const char *)email + ">";

    // make sure we dont have old sources around...
    erasefiles(prefix, str(tarbase) + "-", ".src.rpm");

    argv[index++] = (char *)"rpmbuild";
    argv[index++] = (char *)"-ts";
    argv[index++] = (char *)"--nodeps";
    argv[index++] = (char *)"--define";
    argv[index++] = (char *)*packager;
    argv[index++] = (char *)"--define";
    argv[index++] = (char *)*srcdir;

    if(vendor && *vendor) {
        String v = str("_vendor ") + vendor;
        argv[index++] = (char *)"--define";
        argv[index++] = (char *)*v;
    }

    if(is(sign)) {
        argv[index++] = (char *)"--sign";
        if(signid) {
            String signer = str("_gpg_name ") + (const char *)realname + " <" + (const char *)signid + ">";
            argv[index++] = (char *)"--define";
            argv[index++] = (char *)*signer;
        }
    }

    if(is(verbose))
        argv[index++] = (char *)"-v";
    argv[index++] = (char *)tarball;
    argv[index] = NULL;

    ::exit(fsys::exec(*argv, argv));
}

static void rpm_source(const char *specfile, const char *srcdir = NULL)
{
    check_spec(specfile);

    char *argv[20];
    unsigned index = 0;
    char buffer[512];

    // make sure we dont have old sources around...

    if(srcdir) {
        String from = tarball;
        String to = str(srcdir) + "/" + tarfile;
        erasefiles(srcdir, str(tarbase) + "-", ".tar.gz");
        copyfile(from, to);
    }
    else {
        String::set(buffer, sizeof(buffer), tarball);
        char *cp = strrchr(buffer, '/');
        if(cp) {
            *cp = 0;
            srcdir = buffer;
        }
        else
            srcdir = prefix;
    }

    erasefiles(prefix, str(tarbase) + "-", ".src.rpm");

    String rpmdir = str("_srcrpmdir ") + (const char *)prefix;
    String source = str("_sourcedir ") + (const char *)srcdir;
    String packager = str("_packager ") + (const char *)realname + " <" + (const char *)email + ">";

    argv[index++] = (char *)"rpmbuild";
    argv[index++] = (char *)"-bs";
    argv[index++] = (char *)"--nodeps";
    argv[index++] = (char *)"--define";
    argv[index++] = (char *)*packager;
    argv[index++] = (char *)"--define";
    argv[index++] = (char *)*source;
    argv[index++] = (char *)"--define";
    argv[index++] = (char *)*rpmdir;

    if(vendor && *vendor) {
        String v = str("_vendor ") + vendor;
        argv[index++] = (char *)"--define";
        argv[index++] = (char *)*v;
    }

    if(is(sign)) {
        argv[index++] = (char *)"--sign";
        if(signid) {
            String signer = str("_gpg_name ") + (const char *)realname + " <" + (const char *)signid + ">";
            argv[index++] = (char *)"--define";
            argv[index++] = (char *)*signer;
        }
    }

    if(is(verbose))
        argv[index++] = (char *)"-v";

    argv[index++] = (char *)specfile;
    argv[index] = NULL;

    ::exit(fsys::exec(*argv, argv));
}

static int clean(const char *dir)
{
    if(!fsys::is_dir(dir))
        return -1;

    if(*dir == '/')
        return -1;

    char *argv[] = {(char *)"rm", (char *)"-r", (char *)dir, NULL};
    return fsys::exec(*argv, argv);
}

static void cleanexit(void)
{
    if(pid == getpid() && cleandir && fsys::is_dir(cleandir))
        clean(cleandir);
}

static int tar(const char *tarfile, const char *dir)
{
    const char *argv[] = {"tar", "-czf", tarfile, dir, NULL};
    return fsys::exec(*argv, (char **)argv);
}

static int untar(const char *file)
{
    const char *extract = "-xf";

    if(eq(format, "tar.gz"))
        extract = "-xzf";

    char *argv[] = {(char *)"tar", (char *)extract, (char *)file, NULL};

    return fsys::exec(*argv, argv);
}

static void deb_source(const char *debroot)
{
    String basename;
    String source = str(tarbase) + "-" + version;
    String original = str(tarbase) + "_" + version + ".orig.tar.gz";

    cleandir = strdup(*source);
    atexit(cleanexit);
    clean(*source);

    check_control(str(debroot) + "/debian/control");
    check_changes(str(debroot) + "/debian/changelog");

    if(untar(tarball)) {
        shell::errexit(4, "*** cape-source: %s: %s\n",
            tarball, _TEXT("failed to unpack"));
    }

    if(!fsys::is_dir(*source)) {
        shell::errexit(4, "*** cape-source: %s: %s\n",
            tarball, _TEXT("failed to unpack"));
    }

    // make sure we dont have old sources around...
    basename = str(tarbase) + "_";
    erasefiles(prefix, basename, ".gz");
    erasefiles(prefix, basename, ".dsc");
    erasefiles(prefix, basename, "_source.changes");

    // debianize source...
    if(tar(*original, *source)) {
        shell::errexit(5, "*** cape-source: %s: %s\n",
            *original, _TEXT("cannot create"));
    }

    dir::create(str(source) + "/debian", 0664);
    if(system(str("cp -a ") + debroot + "/debian/* " + source + "/debian"))
        shell::errexit(5, "*** cape-source: %s/*: %s",
            debroot, _TEXT("cannot copy"));

    char *argv[32];
    unsigned index = 0;

    argv[index++] = (char *)"dpkg-source";
    argv[index++] = (char *)"-b";
    argv[index++] = (char *)*source;
    argv[index++] = NULL;

    int status = fsys::exec(*argv, argv);
    if(status)
        exit(status);

    String dsc;

    if(revision)
        dsc = str(tarbase) + "_" + version + "-" + revision + ".dsc";
    else
        dsc = str(tarbase) + "_" + version + ".dsc";

    if(deb_lintian(*dsc))
        shell::errexit(6, "*** cape-source: %s: %s\n",
            *dsc, _TEXT("failed lint\n"));

    if(is(changes)) {
        index = 0;
        argv[index++] = (char *)"dpkg-genchanges";
        argv[index++] = (char *)"-S";
        argv[index++] = NULL;

        String changefile;

        if(revision)
            changefile = str(tarbase) + "_" + version + "-" + revision + "_source.changes";
        else
            changefile = str(tarbase) + "_" + version + "_source.changes";

        pid = fork();
        if(!pid) {
            int fd = open(changefile, O_WRONLY | O_TRUNC | O_CREAT, 0644);
            if(fd < 1)
                shell::errexit(11, "*** cape-source: %s: %s",
                    *changefile, _TEXT("cannot write"));
            dup2(fd, 1);
            if(chdir(*source))
                shell::errexit(10, "*** cape-source: %s: %s",
                    *source, _TEXT("invalid directory"));
            execvp(*argv, argv);
            exit(-1);
        }
        shell::wait(pid);
    }

    clean(*source);

    String gpg;
    if(signid)
        gpg = str("gpg -u ") + signid;
    else
        gpg = str("gpg");

    if(is(sign)) {
        remove(".tmp.sig");
        // bizzareness where empty line before signature...
        fsys::erase(".tmp.out");
        if(system("cp *.dsc .tmp.out"))
            goto failed;
        if(system("echo >>.tmp.out"))
            goto failed;
        if(system(str(gpg) + " --clearsign --output .tmp.sig .tmp.out"))
failed:
            shell::errexit(8, "*** cape-source: %s: %s\n",
                tarbase, _TEXT("failed to sign"));
        fsys::erase(".tmp.out");
        if(system(str("mv -f .tmp.sig *.dsc")))
            fprintf(stderr, "*** cape-source: %s\n",
                _TEXT("cannot update signature"));
    }

    if(is(sign) && is(changes)) {
        remove(".tmp.sig");
        if(system(str(gpg) + " --clearsign --output .tmp.sig *.changes"))
            shell::errexit(8, "*** cape-source: %s: %s\n",
                tarbase, _TEXT("failed to sign"));
        if(system(str("mv -f .tmp.sig *.changes")))
            fprintf(stderr, "*** cape-source: %s\n",
                _TEXT("cannot update signature"));
    }
}

static void deb_arch(void)
{
    char buffer[80];
    FILE *fp = popen("dpkg --print-architecture", "r");
    if(!fp || !fgets(buffer, sizeof(buffer), fp)) {
        if(fp)
            pclose(fp);
        shell::errexit(20, "*** cape-source: %s",
            _TEXT("cannot use dpkg"));
    }
    pclose(fp);
    char *cp = strpbrk(buffer, "\r\n");
    if(cp)
        *cp = 0;
    arch = strdup(buffer);
}

static void rpm_arch(void)
{
    char buffer[80];
    FILE *fp = popen("rpm --eval %_arch", "r");
    if(!fp || !fgets(buffer, sizeof(buffer), fp)) {
        if(fp)
            pclose(fp);
        shell::errexit(20, "*** cape-source: %s",
            _TEXT("cannot use rpm"));
    }
    pclose(fp);
    char *cp = strpbrk(buffer, "\r\n");
    if(cp)
        *cp = 0;
    arch = strdup(buffer);
}

static void rpm_release(void)
{
    char dist[80];
    char dver[80];

    FILE *fp = popen("lsb_release -ir", "r");
    if(!fp || !fgets(dist, sizeof(dist), fp) || !fgets(dver, sizeof(dist), fp)) {
        if(fp)
            pclose(fp);
        shell::errexit(20, "*** cape-source: %s",
            _TEXT("cannot use rpm"));
    }
    pclose(fp);
    char *id = strpbrk(dist, "\r\n");
    if(id)
        *id = 0;

    id = strchr(dist, ':');
    while(id && isspace(*++id));

    char *v = strpbrk(dver, "\r\n");
    if(v)
        *v = 0;

    v = strchr(dver, ':');
    while(v && isspace(*++v));

    if(!id || !v)
        return;

    String::lower(id);
    releaseid = strdup(str(id) + "-" + (const char *)v);
}

static void deb_release(void)
{
    char buffer[80];
    FILE *fp = popen("lsb_release -c", "r");
    if(!fp || !fgets(buffer, sizeof(buffer), fp)) {
        if(fp)
            pclose(fp);
        shell::errexit(20, "*** cape-source: %s",
            _TEXT("cannot use lsb_release"));
    }
    pclose(fp);
    char *cp = strpbrk(buffer, "\r\n");
    if(cp)
        *cp = 0;
    cp = strchr(buffer, ':');
    while(cp && isspace(*++cp));
    if(cp)
        releaseid = strdup(cp);
}

PROGRAM_MAIN(argc, argv)
{
    shell::bind("cape-source");
    shell args(argc, argv);
    argv = args.argv();
    char filename[65];
    char hostname[256];
    char *cp;

    struct passwd *pwd = getpwuid(getuid());

    if(fsys::is_file("/etc/debian_version"))
        debianos = true;

    if(fsys::is_dir("/etc/yum"))
        redhatos = true;

    pid = getpid();

    pkgbuildrc.load(shell::path(shell::USER_HOME) + "/.pkgbuildrc");
    archives = pkgbuildrc.get(*profile);
    gethostname(hostname, sizeof(hostname));

    if(pwd) {
        email = strdup(str(pwd->pw_name) + "@" + (const char *)hostname);
        realname = strdup(pwd->pw_gecos);
        char *ep = strchr((char *)realname, ',');
        if(ep)
            *ep = 0;
    }
    else {
        email = strdup(str(getlogin()) + "@" + (const char *)hostname);
        realname = strdup(getlogin());
    }

    if(fsys::is_file("/etc/redhat_release")) {
        rpm_arch();
        rpm_release();
    }
    else {
        deb_arch();
        deb_release();
    }

    if(archives) {
        label = archives->get("label");
        origin = archives->get("origin");
        vendor = archives->get("vendor");
        distribution = archives->get("distribution");
        const char *em = archives->get("email");
        signid = archives->get("key");
        if(!signid)
            signid = em;
        if(em)
            email = em;

        if(!vendor || !*vendor)
            vendor = label;

        if(!vendor || !*vendor)
            vendor = distribution;
    }

    setenv("DEBEMAIL", email, 1);
    setenv("DEBFULLNAME", realname, 1);

    if(is(who)) {
        if(is(profile))
            printf("profile: %s\n", *profile);
        if(origin && *origin)
            printf("origin:  %s\n", origin);
        printf("name:    %s\n", realname);
        printf("email:   %s\n", email);
        if(signid)
            printf("sign-as: %s\n", signid);
        if(arch)
            printf("arch:    %s\n", arch);
        if(releaseid)
            printf("distro:  %s\n", releaseid);
        exit(0);
    }

    fsys::prefix(prefix, sizeof(prefix));

    if(is(helpflag) || is(althelp)) {
        printf("%s: cape-source %s\n", _TEXT("Usage"),
            _TEXT("[options] tarball [spec]"));
        printf("%s\n\n", _TEXT("Create source package"));
        printf("%s\n", _TEXT("Options:"));
        shell::help();
        printf("\n%s\n", _TEXT("Report bugs to dyfet@gnu.org"));
        PROGRAM_EXIT(0);
    }

    if(!*argv) {
        shell::errexit(1, "*** cape-source: %s\n",
            _TEXT("no tarball to build"));
    }

    tarball = *(argv++);
    tarfile = strrchr(tarball, '/');
    if(tarfile)
        ++tarfile;
    else
        tarfile = tarball;

    String::set(filename, sizeof(filename), tarfile);
    cp = filename;
    while(NULL != (cp = strchr(++cp, '.'))) {
        if(eq(cp, ".tar.gz"))
            break;
    }

    if(!cp)
        shell::errexit(2, "*** cape-source: %s: %s\n",
            filename, _TEXT("unknown source format"));

    *(cp++) = 0;
    format = cp;

    cp = filename;
    tarbase = filename;

    bool dash = false;
    while(*(++cp)) {
        if(*cp == '-') {
            dash = true;
            continue;
        }
        if(dash && *cp >= '0' && *cp <= '9') {
            --cp;
            *cp = 0;
            version = ++cp;
            break;
        }
        dash = false;
    }

    if(!version) {
        shell::errexit(2, "*** cape-source: %s\n",
            _TEXT("cannot detect version"));
    }

    if(is(verbose))
        printf("packaging %s version %s\n", tarbase, version);

    char *spec = *argv;

    // if pure tarball, then rpm with internal spec...
    if(!spec)
        rpm_tarball();

    if(fsys::is_dir(spec)) {
        String specfile = str(spec) + "/" + tarbase + ".spec";
        String debian = str(spec) + "/debian";
        if(fsys::is_dir(*debian))
            deb_source(spec);
        else if(fsys::is_file(*specfile))
            rpm_source(*specfile, spec);
        else
            shell::errexit(3, "*** cape-source %s: %s\n",
                *argv, _TEXT("no spec or debian directory"));
    } else {
        cp = strrchr(spec, '.');
        if(eq(cp, ".spec"))
            rpm_source(spec);
        shell::errexit(3, "*** cape-source: %s: %s\n",
            spec, _TEXT("not rpm spec file"));
    }

    PROGRAM_EXIT(0);
}


